"""A wrapper for numbers."""
from __future__ import annotations

from typing import Any

from ee import _arg_types
from ee import _cloud_api_utils
from ee import _utils
from ee import apifunction
from ee import computedobject
from ee import ee_exception
from ee import ee_string


class Number(computedobject.ComputedObject):
  """An object to represent numbers."""

  _number: float | None

  _initialized = False

  # Tell pytype to not complain about dynamic attributes.
  _HAS_DYNAMIC_ATTRIBUTES = True

  def __init__(self, number: _arg_types.Number):
    """Construct a number wrapper.

    This constructor accepts the following args:
      1) A bare number.
      2) A ComputedObject returning a number.

    Args:
      number: The number to wrap.
    """
    self.initialize()

    if isinstance(number, (float, int)):
      super().__init__(None, None)
      self._number = number
    elif isinstance(number, computedobject.ComputedObject):
      super().__init__(number.func, number.args, number.varName)
      self._number = None
    else:
      raise ee_exception.EEException(
          'Invalid argument specified for ee.Number(): %s' % number)

  @classmethod
  def initialize(cls) -> None:
    """Imports API functions to this class."""
    if not cls._initialized:
      apifunction.ApiFunction.importApi(cls, cls.name(), cls.name())
      cls._initialized = True

  @classmethod
  def reset(cls) -> None:
    """Removes imported API functions from this class."""
    apifunction.ApiFunction.clearApi(cls)
    cls._initialized = False

  @staticmethod
  def name() -> str:
    return 'Number'

  @_utils.accept_opt_prefix('opt_encoder')
  def encode(self, encoder: Any = None) -> Any:
    if isinstance(self._number, (float, int)):
      return self._number
    else:
      return super().encode(encoder)

  @_utils.accept_opt_prefix('opt_encoder')
  def encode_cloud_value(self, encoder: Any = None) -> Any:
    if isinstance(self._number, (float, int)):
      return _cloud_api_utils.encode_number_as_cloud_value(self._number)
    else:
      return super().encode_cloud_value(encoder)

  def abs(self) -> Number:
    """Computes the absolute value of the input."""

    return apifunction.ApiFunction.call_(self.name() + '.abs', self)

  def acos(self) -> Number:
    """Computes the arccosine in radians of the input."""

    return apifunction.ApiFunction.call_(self.name() + '.acos', self)

  def add(self, right: _arg_types.Number) -> Number:
    """Adds the right value.

    Args:
      right: The value to add to the current ee.Number.

    Returns:
      An ee.Number.
    """

    return apifunction.ApiFunction.call_(self.name() + '.add', self, right)

  # `and` is not allowed by the Python parser.
  def And(self, right: _arg_types.Number) -> Number:
    """Returns 1 if and only if both values are non-zero.

    Args:
      right: The value to and with the current ee.Number.

    Returns:
      An ee.Number.
    """

    return apifunction.ApiFunction.call_(self.name() + '.and', self, right)

  def asin(self) -> Number:
    """Computes the arcsine in radians of the input."""

    return apifunction.ApiFunction.call_(self.name() + '.asin', self)

  def atan(self) -> Number:
    """Computes the arctangent in radians of the input."""

    return apifunction.ApiFunction.call_(self.name() + '.atan', self)

  def atan2(self, right: _arg_types.Number) -> Number:
    """Calculates the angle in radians formed by the 2D vector [x, y].

    Args:
      right: The second value of the atan2 call.

    Returns:
      An ee.Number.
    """

    return apifunction.ApiFunction.call_(self.name() + '.atan2', self, right)

  def bitCount(self) -> Number:
    """Returns the number of one-bits in the Number.

    Calculates the number of one-bits in the 64-bit two's complement binary
    representation of the input.

    Returns:
      An ee.Number.
    """

    return apifunction.ApiFunction.call_(self.name() + '.bitCount', self)

  def bitwiseAnd(self, right: _arg_types.Number) -> Number:
    """Calculates the bitwise AND of the input values.

    Args:
      right: The value to do the bitwise AND with.

    Returns:
      An ee.Number.
    """

    return apifunction.ApiFunction.call_(
        self.name() + '.bitwiseAnd', self, right
    )

  def bitwiseNot(self) -> Number:
    """Returns the bitwise NOT of the input.

    Uses the smallest signed integer type that can hold the input.
    """

    return apifunction.ApiFunction.call_(self.name() + '.bitwiseNot', self)

  def bitwiseOr(self, right: _arg_types.Number) -> Number:
    """Calculates the bitwise OR of the input values.

    Args:
      right: The value to OR with.

    Returns:
      An ee.Number.
    """

    return apifunction.ApiFunction.call_(
        self.name() + '.bitwiseOr', self, right
    )

  def bitwiseXor(self, right: _arg_types.Number) -> Number:
    """Calculates the bitwise XOR of the input values.

    Args:
      right: The right-hand value.

    Returns:
      An ee.Number.
    """

    return apifunction.ApiFunction.call_(
        self.name() + '.bitwiseXor', self, right
    )

  def byte(self) -> Number:
    """Casts the input value to an unsigned 8-bit integer."""

    return apifunction.ApiFunction.call_(self.name() + '.byte', self)

  def cbrt(self) -> Number:
    """Computes the cubic root of the input."""

    return apifunction.ApiFunction.call_(self.name() + '.cbrt', self)

  def ceil(self) -> Number:
    """Computes the smallest integer greater than or equal to the input."""

    return apifunction.ApiFunction.call_(self.name() + '.ceil', self)

  def clamp(
      self,
      min: _arg_types.Number,  # pylint: disable=redefined-builtin
      max: _arg_types.Number,  # pylint: disable=redefined-builtin
  ) -> Number:
    """Returns a number that is clamped to lie within the range of min to max.

    Args:
      min: The minimum value to clamp to.
      max: The maximum value to clamp to.

    Returns:
      An ee.Number.
    """

    return apifunction.ApiFunction.call_(self.name() + '.clamp', self, min, max)

  def cos(self) -> Number:
    """Computes the cosine of the input in radians."""

    return apifunction.ApiFunction.call_(self.name() + '.cos', self)

  def cosh(self) -> Number:
    """Computes the hyperbolic cosine of the input."""

    return apifunction.ApiFunction.call_(self.name() + '.cosh', self)

  def digamma(self) -> Number:
    """Computes the digamma function of the input."""

    return apifunction.ApiFunction.call_(self.name() + '.digamma', self)

  def divide(self, right: _arg_types.Number) -> Number:
    """Divides the first value by the second, returning 0 for division by 0.

    Args:
      right: The value to divide by.

    Returns:
      An ee.Number.
    """

    return apifunction.ApiFunction.call_(self.name() + '.divide', self, right)

  def double(self) -> Number:
    """Casts the input value to a 64-bit float."""

    return apifunction.ApiFunction.call_(self.name() + '.double', self)

  def eq(self, right: _arg_types.Number) -> Number:
    """Returns 1 if and only if the first value is equal to the second.

    Args:
      right: The value to compare to.

    Returns:
      An ee.Number.
    """

    return apifunction.ApiFunction.call_(self.name() + '.eq', self, right)

  def erf(self) -> Number:
    """Computes the error function of the input."""

    return apifunction.ApiFunction.call_(self.name() + '.erf', self)

  def erfInv(self) -> Number:
    """Computes the inverse error function of the input."""

    return apifunction.ApiFunction.call_(self.name() + '.erfInv', self)

  def erfc(self) -> Number:
    """Computes the complementary error function of the input."""

    return apifunction.ApiFunction.call_(self.name() + '.erfc', self)

  def erfcInv(self) -> Number:
    """Computes the inverse complementary error function of the input."""

    return apifunction.ApiFunction.call_(self.name() + '.erfcInv', self)

  def exp(self) -> Number:
    """Computes the Euler's number e raised to the power of the input."""

    return apifunction.ApiFunction.call_(self.name() + '.exp', self)

  @staticmethod
  def expression(
      expression: _arg_types.String,
      # pylint: disable-next=redefined-builtin
      vars: _arg_types.Dictionary | None = None,
  ) -> Number:
    """Returns a number from computing a numeric expression.

    Args:
      expression: A mathematical expression string to be evaluated. In addition
        to the standard arithmetic, boolean and relational operators,
        expressions also support any function in Number, the '.' operator to
        extract child elements from the 'vars' dictionary, and mathematical
        constants math.pi and math.e.
      vars: A dictionary of named values that can be used in the expression.
    """

    return apifunction.ApiFunction.call_('Number.expression', expression, vars)

  def first(self, right: _arg_types.Number) -> Number:
    """Selects the value of the first value.

    Args:
      right: This value is never returned.

    Returns:
      Return an ee.Number.
    """

    return apifunction.ApiFunction.call_(self.name() + '.first', self, right)

  def firstNonZero(self, right: _arg_types.Number) -> Number:
    """Returns the first value if it is non-zero, otherwise the second value.

    Args:
      right: The second value.

    Returns:
      An ee.Number.
    """

    return apifunction.ApiFunction.call_(
        self.name() + '.firstNonZero', self, right
    )

  def float(self) -> Number:
    """Casts the input value to a 32-bit float."""

    return apifunction.ApiFunction.call_(self.name() + '.float', self)

  def floor(self) -> Number:
    """Computes the largest integer less than or equal to the input."""

    return apifunction.ApiFunction.call_(self.name() + '.floor', self)

  def format(
      self, pattern: _arg_types.String | None = None
  ) -> ee_string.String:
    r"""Convert a number to a string using printf-style formatting.

    For more about format strings, see

      https://docs.oracle.com/en/java/javase/11/docs/api/java.base/java/util/Formatter.html

    Args:
      pattern: A printf-style format string. For example, '%.2f' produces
        numbers formatted like '3.14', and '%05d' produces numbers formatted
        like '00042'. The format string must satisfy the following criteria:

        1. Zero or more prefix characters.
        2. Exactly one '%'.
        3. Zero or more modifier characters in the set [#-+ 0,(.\d].
        4. Exactly one conversion character in the set [sdoxXeEfgGaA].
        5. Zero or more suffix characters.

    Returns:
      An ee.String with the number formatted based on the pattern.
    """  # fmt: skip

    return apifunction.ApiFunction.call_(self.name() + '.format', self, pattern)

  def gamma(self) -> Number:
    """Computes the gamma function of the input."""

    return apifunction.ApiFunction.call_(self.name() + '.gamma', self)

  def gammainc(self, right: _arg_types.Number) -> Number:
    """Calculates the regularized lower incomplete Gamma function γ(x,a).

    Args:
      right: The right-hand value.

    Returns:
      An ee.Number.
    """

    return apifunction.ApiFunction.call_(self.name() + '.gammainc', self, right)

  def gt(self, right: _arg_types.Number) -> Number:
    """Returns 1 if and only if the first value is greater than the second.

    Args:
      right: The right-hand value.

    Returns:
      An ee.Number.
    """

    return apifunction.ApiFunction.call_(self.name() + '.gt', self, right)

  def gte(self, right: _arg_types.Number) -> Number:
    """Returns 1 if the first value is greater than or equal to the second.

    Args:
      right: The right-hand value.

    Returns:
      An ee.Number.
    """

    return apifunction.ApiFunction.call_(self.name() + '.gte', self, right)

  def hypot(self, right: _arg_types.Number) -> Number:
    """Calculates the magnitude of the 2D vector [x, y].

    Args:
      right: The y value.

    Returns:
      An ee.Number.
    """

    return apifunction.ApiFunction.call_(self.name() + '.hypot', self, right)

  def int(self) -> Number:
    """Casts the input value to a signed 32-bit integer."""

    return apifunction.ApiFunction.call_(self.name() + '.int', self)

  def int16(self) -> Number:
    """Casts the input value to a signed 16-bit integer."""

    return apifunction.ApiFunction.call_(self.name() + '.int16', self)

  def int32(self) -> Number:
    """Casts the input value to a signed 32-bit integer."""

    return apifunction.ApiFunction.call_(self.name() + '.int32', self)

  def int64(self) -> Number:
    """Casts the input value to a signed 64-bit integer."""

    return apifunction.ApiFunction.call_(self.name() + '.int64', self)

  def int8(self) -> Number:
    """Casts the input value to a signed 8-bit integer."""

    return apifunction.ApiFunction.call_(self.name() + '.int8', self)

  def lanczos(self) -> Number:
    """Computes the Lanczos approximation of the input."""

    return apifunction.ApiFunction.call_(self.name() + '.lanczos', self)

  def leftShift(self, right: _arg_types.Number) -> Number:
    """Calculates the left shift of v1 by v2 bits.

    Args:
      right: How many bits to shift.

    Returns:
      An ee.Number.
    """

    return apifunction.ApiFunction.call_(
        self.name() + '.leftShift', self, right
    )

  def log(self) -> Number:
    """Computes the natural logarithm of the input."""

    return apifunction.ApiFunction.call_(self.name() + '.log', self)

  def log10(self) -> Number:
    """Computes the base-10 logarithm of the input."""

    return apifunction.ApiFunction.call_(self.name() + '.log10', self)

  def long(self) -> Number:
    """Casts the input value to a signed 64-bit integer."""

    return apifunction.ApiFunction.call_(self.name() + '.long', self)

  def lt(self, right: _arg_types.Number) -> Number:
    """Returns 1 if and only if the first value is less than the second.

    Args:
      right: The value to compare to.

    Returns:
      An ee.Number.
    """

    return apifunction.ApiFunction.call_(self.name() + '.lt', self, right)

  def lte(self, right: _arg_types.Number) -> Number:
    """Returns 1 if the first value is less than or equal to the second.

    Args:
      right: The value to compare to.

    Returns:
      An ee.Number.
    """

    return apifunction.ApiFunction.call_(self.name() + '.lte', self, right)

  def max(self, right: _arg_types.Number) -> Number:
    """Selects the maximum of the first and second values.

    Args:
      right: The value to compare to.

    Returns:
      An ee.Number.
    """

    return apifunction.ApiFunction.call_(self.name() + '.max', self, right)

  def min(self, right: _arg_types.Number) -> Number:
    """Selects the minimum of the first and second values.

    Args:
      right: The value to compare to.

    Returns:
      An ee.Number.
    """

    return apifunction.ApiFunction.call_(self.name() + '.min', self, right)

  def mod(self, right: _arg_types.Number) -> Number:
    """Calculates the remainder of the first value divided by the second.

    Args:
      right: The value to divide by.

    Returns:
      An ee.Number.
    """

    return apifunction.ApiFunction.call_(self.name() + '.mod', self, right)

  def multiply(self, right: _arg_types.Number) -> Number:
    """Multiplies the first value by the second.

    Args:
      right: The value to multiply by.

    Returns:
      An ee.Number.
    """

    return apifunction.ApiFunction.call_(self.name() + '.multiply', self, right)

  def neq(self, right: _arg_types.Number) -> Number:
    """Returns 1 if and only if the first value is not equal to the second.

    Args:
      right: The value to compare to.

    Returns:
      An ee.Number.
    """

    return apifunction.ApiFunction.call_(self.name() + '.neq', self, right)

  def Not(self) -> Number:
    """Returns 0 if the input is non-zero, and 1 otherwise."""

    return apifunction.ApiFunction.call_(self.name() + '.not', self)

  def Or(self, right: _arg_types.Number) -> Number:
    """Returns 1 if and only if either input value is non-zero.

    Args:
      right: The value to or with.
    """

    return apifunction.ApiFunction.call_(self.name() + '.or', self, right)

  @staticmethod
  def parse(
      input: _arg_types.String,  # pylint: disable=redefined-builtin
      radix: _arg_types.Integer | None = None,
  ) -> Number:
    """Returns a number from a string.

    Args:
      input: The string to convert to a number.
      radix: An integer representing the base number system from which to
        convert. If input is not an integer, radix must equal 10 or not be
        specified.
    """

    return apifunction.ApiFunction.call_('Number.parse', input, radix)

  def pow(self, right: _arg_types.Number) -> Number:
    """Raises the first value to the power of the second.

    Args:
      right: The exponent to raise the value to.

    Returns:
      An ee.Number.
    """

    return apifunction.ApiFunction.call_(self.name() + '.pow', self, right)

  def rightShift(self, right: _arg_types.Number) -> Number:
    """Calculates the signed right shift of v1 by v2 bits.

    Args:
      right: How many bits to shift.

    Returns:
      An ee.Number.
    """

    return apifunction.ApiFunction.call_(
        self.name() + '.rightShift', self, right
    )

  def round(self) -> Number:
    """Computes the integer nearest to the input."""

    return apifunction.ApiFunction.call_(self.name() + '.round', self)

  def short(self) -> Number:
    """Casts the input value to a signed 16-bit integer."""

    return apifunction.ApiFunction.call_(self.name() + '.short', self)

  def signum(self) -> Number:
    """Computes the signum function (sign) of the input.

    The return value is 0 if the input is 0, 1 if the input is greater than 0,
    -1 if the input is less than 0.

    Returns:
      An ee.Number.
    """

    return apifunction.ApiFunction.call_(self.name() + '.signum', self)

  def sin(self) -> Number:
    """Computes the sine of the input in radians."""

    return apifunction.ApiFunction.call_(self.name() + '.sin', self)

  def sinh(self) -> Number:
    """Computes the hyperbolic sine of the input."""

    return apifunction.ApiFunction.call_(self.name() + '.sinh', self)

  def sqrt(self) -> Number:
    """Computes the square root of the input."""

    return apifunction.ApiFunction.call_(self.name() + '.sqrt', self)

  def subtract(self, right: _arg_types.Number) -> Number:
    """Subtracts the second value from the first.

    Args:
      right: The value to subtract.

    Returns:
      An ee.Number.
    """

    return apifunction.ApiFunction.call_(self.name() + '.subtract', self, right)

  def tan(self) -> Number:
    """Computes the tangent of the input in radians."""

    return apifunction.ApiFunction.call_(self.name() + '.tan', self)

  def tanh(self) -> Number:
    """Computes the hyperbolic tangent of the input."""

    return apifunction.ApiFunction.call_(self.name() + '.tanh', self)

  def toByte(self) -> Number:
    """Casts the input value to an unsigned 8-bit integer."""

    return apifunction.ApiFunction.call_(self.name() + '.toByte', self)

  def toDouble(self) -> Number:
    """Casts the input value to a 64-bit float."""

    return apifunction.ApiFunction.call_(self.name() + '.toDouble', self)

  def toFloat(self) -> Number:
    """Casts the input value to a 32-bit float."""

    return apifunction.ApiFunction.call_(self.name() + '.toFloat', self)

  def toInt(self) -> Number:
    """Casts the input value to a signed 32-bit integer."""

    return apifunction.ApiFunction.call_(self.name() + '.toInt', self)

  def toInt16(self) -> Number:
    """Casts the input value to a signed 16-bit integer."""

    return apifunction.ApiFunction.call_(self.name() + '.toInt16', self)

  def toInt32(self) -> Number:
    """Casts the input value to a signed 32-bit integer."""

    return apifunction.ApiFunction.call_(self.name() + '.toInt32', self)

  def toInt64(self) -> Number:
    """Casts the input value to a signed 64-bit integer."""

    return apifunction.ApiFunction.call_(self.name() + '.toInt64', self)

  def toInt8(self) -> Number:
    """Casts the input value to a signed 8-bit integer."""

    return apifunction.ApiFunction.call_(self.name() + '.toInt8', self)

  def toLong(self) -> Number:
    """Casts the input value to a signed 64-bit integer."""

    return apifunction.ApiFunction.call_(self.name() + '.toLong', self)

  def toShort(self) -> Number:
    """Casts the input value to a signed 16-bit integer."""

    return apifunction.ApiFunction.call_(self.name() + '.toShort', self)

  def toUint16(self) -> Number:
    """Casts the input value to an unsigned 16-bit integer."""

    return apifunction.ApiFunction.call_(self.name() + '.toUint16', self)

  def toUint32(self) -> Number:
    """Casts the input value to an unsigned 32-bit integer."""

    return apifunction.ApiFunction.call_(self.name() + '.toUint32', self)

  def toUint8(self) -> Number:
    """Casts the input value to an unsigned 8-bit integer."""

    return apifunction.ApiFunction.call_(self.name() + '.toUint8', self)

  def trigamma(self) -> Number:
    """Computes the trigamma function of the input."""

    return apifunction.ApiFunction.call_(self.name() + '.trigamma', self)

  def uint16(self) -> Number:
    """Casts the input value to an unsigned 16-bit integer."""

    return apifunction.ApiFunction.call_(self.name() + '.uint16', self)

  def uint32(self) -> Number:
    """Casts the input value to an unsigned 32-bit integer."""

    return apifunction.ApiFunction.call_(self.name() + '.uint32', self)

  def uint8(self) -> Number:
    """Casts the input value to an unsigned 8-bit integer."""

    return apifunction.ApiFunction.call_(self.name() + '.uint8', self)

  # pylint: disable-next=redefined-builtin
  def unitScale(self, min: _arg_types.Number, max: _arg_types.Number) -> Number:
    """Returns the input scaled so that [min, max] becomes [0, 1].

    Values outside the range are NOT clamped. If min == max, 0 is returned.

    Args:
      min: Minimum value of the input to be scaled to 0.
      max: Maximum value of the input to be scaled to 1.
    """

    return apifunction.ApiFunction.call_(
        self.name() + '.unitScale', self, min, max
    )



#!/usr/bin/env python3
"""Tests for ee._state."""

from absl.testing import parameterized

import unittest
from ee import _state


class StateTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    _state.reset_state()

  @parameterized.named_parameters(
      ('global_mode', False),
  )
  def test_get_state(self, use_context_mode: bool):
    state = _state.get_state()

    self.assertIsInstance(state, _state.EEState)
    self.assertEqual(state, _state.EEState())

  @parameterized.named_parameters(
      ('global_mode', False),
  )
  def test_update_state(self, use_context_mode: bool):
    state = _state.get_state()

    # Modify the state and verify the global state has been updated.
    state.cloud_api_user_project = 'my-project'

    self.assertEqual(_state.get_state().cloud_api_user_project, 'my-project')

  @parameterized.named_parameters(
      ('global_mode', False),
  )
  def test_reset_state(self, use_context_mode: bool):
    state = _state.get_state()
    state.cloud_api_user_project = 'my-project'

    _state.reset_state()

    self.assertEqual(_state.get_state(), _state.EEState())


if __name__ == '__main__':
  unittest.main()



"""A wrapper for Arrays."""
from __future__ import annotations

from typing import Any

from ee import _arg_types
from ee import apifunction
from ee import computedobject
from ee import dictionary
from ee import ee_list
from ee import ee_number


class Array(computedobject.ComputedObject):
  """An object to represent an Earth Engine Array.

  Examples:
    # Requires an explicit PixelType if no data.
    ee.Array([], ee.PixelType.int8()); value: []
    ee.Array([[]], ee.PixelType.uint8()); value: [[]]
    ee.Array([[], []], ee.PixelType.float()); value: [[], []]
    # 1-D Arrays
    ee.Array([0]); value: [0]
    ee.Array([0, 1]); value: [0, 1]
    # 2-D Arrays
    ee.Array([[1]]); value: [[1]]
    ee.Array([[0, 1], [2, 3]]); value: [[0,1],[2,3]]
    # Arrays from ee.Number.
    ee.Array([ee.Number(123).toUint8()]); value: [123]
    # Lists are useful ways to construct larger Arrays.
    ee.Array(ee.List.sequence(0, 10, 2)); value: [0, 2, 4, 6, 8, 10]
  """

  _initialized: bool = False

  def __init__(
      self,
      values: _arg_types.Array,
      # pylint: disable-next=invalid-name
      pixelType: _arg_types.String | None = None,
  ):
    """Creates an Array wrapper.

    Returns an Array with the given coordinates.

    Args:
      values: An existing array to cast, or a number/list of numbers/nested list
        of numbers of any depth to create an array from. For nested lists, all
        inner arrays at the same depth must have the same length, and numbers
        may only be present at the deepest level.
      pixelType: The type of each number in the values argument. If the pixel
        type is not provided, it will be inferred from the numbers in 'values'.
        If there aren't any numbers in 'values', this type must be provided.
    """
    self.initialize()

    if isinstance(values, computedobject.ComputedObject) and pixelType is None:
      if self.is_func_returning_same(values):
        # If it is a call that is already returning an Array, just cast.
        super().__init__(values.func, values.args, values.varName)
        return

    args: dict[str, Any] = {'values': values}
    if pixelType is not None:
      args['pixelType'] = pixelType

    func = apifunction.ApiFunction(self.name())
    super().__init__(func, func.promoteArgs(args))

  @classmethod
  def initialize(cls) -> None:
    """Imports API functions to this class."""
    if not cls._initialized:
      apifunction.ApiFunction.importApi(cls, cls.name(), cls.name())
      cls._initialized = True

  @classmethod
  def reset(cls) -> None:
    """Removes imported API functions from this class."""
    apifunction.ApiFunction.clearApi(cls)
    cls._initialized = False

  @staticmethod
  def name() -> str:
    return 'Array'

  def abs(self) -> Array:
    """On an element-wise basis, computes the absolute value of the input."""

    return apifunction.ApiFunction.call_(self.name() + '.abs', self)

  def accum(
      self,
      axis: _arg_types.Integer,
      reducer: _arg_types.Reducer | None = None,
  ) -> Array:
    """Accumulates elements of an array along the given axis.

    Sets each element of the result to the reduction of elements along that axis
    up to and including the current position.

    May be used to make a cumulative sum, a monotonically increasing sequence,
    etc.

    Args:
      axis: Axis along which to perform the accumulation.
      reducer: Reducer to accumulate values. Default is SUM, to produce the
        cumulative sum of each vector along the given axis.

    Returns:
      An ee.Array.
    """

    return apifunction.ApiFunction.call_(
        self.name() + '.accum', self, axis, reducer
    )

  def acos(self) -> Array:
    """Computes the arccosine in radians of the input."""

    return apifunction.ApiFunction.call_(self.name() + '.acos', self)

  def add(self, right: _arg_types.Array) -> Array:
    """On an element-wise basis, adds the first value to the second.

    Args:
      right: The values to add.

    Returns:
      An ee.Array.
    """

    return apifunction.ApiFunction.call_(self.name() + '.add', self, right)

  def And(self, right: _arg_types.Array) -> Array:
    """Returns 1 if both values are non-zero.

    Args:
      right: The values to add.

    Returns:
      An ee.Array.
    """

    return apifunction.ApiFunction.call_(self.name() + '.and', self, right)

  def argmax(self) -> ee_list.List:
    """Returns the position of the maximum value in an array.

    If there are multiple occurrences of the maximum, returns the position of
    the first.

    Returns:
      An ee.List that is empty if the input in an empty ee.Array.
    """

    return apifunction.ApiFunction.call_(self.name() + '.argmax', self)

  def asin(self) -> Array:
    """Computes the arcsine in radians of the input."""

    return apifunction.ApiFunction.call_(self.name() + '.asin', self)

  def atan(self) -> Array:
    """Computes the arctangent in radians of the input."""

    return apifunction.ApiFunction.call_(self.name() + '.atan', self)

  def atan2(self, right: _arg_types.Array) -> Array:
    """Calculates the angle formed by the 2D vector [x, y].

    Args:
      right: The y values.

    Returns:
      An ee.Array.
    """

    return apifunction.ApiFunction.call_(self.name() + '.atan2', self, right)

  def bitCount(self) -> Array:
    """Calculates the number of one-bits.

    Uses a 64-bit two's complement binary representation of the input.

    Returns:
      An ee.Array.
    """

    return apifunction.ApiFunction.call_(self.name() + '.bitCount', self)

  @staticmethod
  # pylint: disable-next=redefined-builtin
  def bitsToArray(input: _arg_types.Integer) -> Array:
    """Returns an Array from the bits of an integer.

    The array has as many elements as the position of the highest set bit, or a
    single 0 for a value of 0.

    Args:
      input: The integer to transform.
    """

    return apifunction.ApiFunction.call_('Array.bitsToArray', input)

  def bitwiseAnd(self, right: _arg_types.Array) -> Array:
    """On an element-wise basis, calculates the bitwise AND of the input values.

    Args:
      right: The values to AND with.

    Returns:
      An ee.Array.
    """

    return apifunction.ApiFunction.call_(
        self.name() + '.bitwiseAnd', self, right
    )

  def bitwiseNot(self) -> Array:
    """On an element-wise basis, calculates the bitwise NOT of the input.

    In the smallest signed integer type that can hold the input.

    Returns:
      An ee.Array.
    """

    return apifunction.ApiFunction.call_(self.name() + '.bitwiseNot', self)

  def bitwiseOr(self, right: _arg_types.Array) -> Array:
    """On an element-wise basis, calculates the bitwise OR of the input values.

    Args:
      right: The values to OR with.

    Returns:
      An ee.Array.
    """

    return apifunction.ApiFunction.call_(
        self.name() + '.bitwiseOr', self, right
    )

  def bitwiseXor(self, right: _arg_types.Array) -> Array:
    """On an element-wise basis, calculates the bitwise XOR of the input values.

    Args:
      right: The values to XOR with.

    Returns:
      An ee.Array.
    """

    return apifunction.ApiFunction.call_(
        self.name() + '.bitwiseXor', self, right
    )

  def byte(self) -> Array:
    """Casts the input value to an unsigned 8-bit integer."""

    return apifunction.ApiFunction.call_(self.name() + '.byte', self)

  @staticmethod
  def cat(
      arrays: _arg_types.List, axis: _arg_types.Integer | None = None
  ) -> Array:
    """Returns an Array that is the concatenation of the given arrays.

    Concatenates multiple arrays into a single array along the given axis. Each
    array must have the same dimensionality and the same length on all axes
    except the concatenation axis.

    Args:
      arrays: Arrays to concatenate.
      axis: Axis to concatenate along. Defaults to 0.
    """

    return apifunction.ApiFunction.call_('Array.cat', arrays, axis)

  def cbrt(self) -> Array:
    """On an element-wise basis, computes the cubic root of the input."""

    return apifunction.ApiFunction.call_(self.name() + '.cbrt', self)

  def ceil(self) -> Array:
    """Computes the smallest integer greater than or equal to the input."""

    return apifunction.ApiFunction.call_(self.name() + '.ceil', self)

  def cos(self) -> Array:
    """On an element-wise basis, computes the cosine of the input in radians."""

    return apifunction.ApiFunction.call_(self.name() + '.cos', self)

  def cosh(self) -> Array:
    """On an element-wise basis, computes the hyperbolic cosine of the input."""

    return apifunction.ApiFunction.call_(self.name() + '.cosh', self)

  def cut(self, position: _arg_types.List) -> Array:
    """Cut an array along one or more axes.

    Args:
      position: Cut an array along one or more axes. The positions args
        specifies either a single value for each axis of the array, or -1,
        indicating the whole axis. The output will be an array that has the same
        dimensions as the input, with a length of 1 on each axis that was not -1
        in the positions array.

    Returns:
      An ee.Array.
    """

    return apifunction.ApiFunction.call_(self.name() + '.cut', self, position)

  def digamma(self) -> Array:
    """On an element-wise basis, computes the digamma function of the input."""

    return apifunction.ApiFunction.call_(self.name() + '.digamma', self)

  def divide(self, right: _arg_types.Array) -> Array:
    """Divides the first value by the second, returning 0 for division by 0.

    Args:
      right: The values to divide by.

    Returns:
      An ee.Array.
    """

    return apifunction.ApiFunction.call_(self.name() + '.divide', self, right)

  def dotProduct(self, array2: _arg_types.Array) -> ee_number.Number:
    """Compute the dot product between two 1-D arrays.

    Args:
      array2: The second 1-D array.

    Returns:
      An ee.Number.
    """

    return apifunction.ApiFunction.call_(
        self.name() + '.dotProduct', self, array2
    )

  def double(self) -> Array:
    """On an element-wise basis, casts the input value to a 64-bit float."""

    return apifunction.ApiFunction.call_(self.name() + '.double', self)

  def eigen(self) -> Array:
    """Computes the real eigenvectors and eigenvalues of a square 2D array.

    Returns:
      An ee.Array with A rows and A+1 columns, where each row contains an
      eigenvalue in the first column, and the corresponding eigenvector in the
      remaining A columns. The rows are sorted by eigenvalue, in descending
      order. This implementation uses DecompositionFactory.eig() from
      http://ejml.org.
    """

    return apifunction.ApiFunction.call_(self.name() + '.eigen', self)

  def eq(self, right: _arg_types.Array) -> Array:
    """Returns 1 if the first value is equal to the second.

    Args:
      right: The values to compare with.

    Returns:
      An ee.Array.
    """

    return apifunction.ApiFunction.call_(self.name() + '.eq', self, right)

  def erf(self) -> Array:
    """On an element-wise basis, computes the error function of the input."""

    return apifunction.ApiFunction.call_(self.name() + '.erf', self)

  def erfInv(self) -> Array:
    """Computes the inverse error function of the input."""

    return apifunction.ApiFunction.call_(self.name() + '.erfInv', self)

  def erfc(self) -> Array:
    """Computes the complementary error function of the input."""

    return apifunction.ApiFunction.call_(self.name() + '.erfc', self)

  def erfcInv(self) -> Array:
    """Computes the inverse complementary error function of the input."""

    return apifunction.ApiFunction.call_(self.name() + '.erfcInv', self)

  def exp(self) -> Array:
    """Computes Euler's number e raised to the power of the input."""

    return apifunction.ApiFunction.call_(self.name() + '.exp', self)

  def first(self, right: _arg_types.Array) -> Array:
    """On an element-wise basis, selects the value of the first value.

    Args:
      right: The matching values that are always ignored.

    Returns:
      An ee.Array.
    """

    return apifunction.ApiFunction.call_(self.name() + '.first', self, right)

  def firstNonZero(self, right: _arg_types.Array) -> Array:
    """Selects the first value if it is non-zero, otherwise the second value.

    Args:
      right: The values to use if the array has a zero in the corresponding
        position.

    Returns:
      An ee.Array.
    """

    return apifunction.ApiFunction.call_(
        self.name() + '.firstNonZero', self, right
    )

  def float(self) -> Array:
    """On an element-wise basis, casts the input value to a 32-bit float."""

    return apifunction.ApiFunction.call_(self.name() + '.float', self)

  def floor(self) -> Array:
    """Computes the largest integer less than or equal to the input."""

    return apifunction.ApiFunction.call_(self.name() + '.floor', self)

  def gamma(self) -> Array:
    """On an element-wise basis, computes the gamma function of the input."""

    return apifunction.ApiFunction.call_(self.name() + '.gamma', self)

  def gammainc(self, right: _arg_types.Any) -> Array:
    """Calculates the regularized lower incomplete Gamma function γ(x,a).

    Args:
      right: The a values for gammainc(x, a).

    Returns:
      An ee.Array.
    """

    return apifunction.ApiFunction.call_(self.name() + '.gammainc', self, right)

  def get(self, position: _arg_types.Any) -> ee_number.Number:
    """Extracts the value at the given position from the input array.

    Args:
      position: The coordinates of the element to get.

    Returns:
      An ee.Number.
    """

    return apifunction.ApiFunction.call_(self.name() + '.get', self, position)

  def gt(self, right: _arg_types.Any) -> Array:
    """Returns 1 if the first value is greater than the second.

    Args:
      right: The values to compare with.

    Returns:
      An ee.Array.
    """

    return apifunction.ApiFunction.call_(self.name() + '.gt', self, right)

  def gte(self, right: _arg_types.Any) -> Array:
    """Returns 1 if the first value is greater than or equal to the second.

    Args:
      right: The values to compare with.

    Returns:
      An ee.Array.
    """

    return apifunction.ApiFunction.call_(self.name() + '.gte', self, right)

  def hypot(self, right: _arg_types.Any) -> Array:
    """Calculates the magnitude of the 2D vector [x, y].

    Args:
      right: The y values.

    Returns:
      An ee.Array.
    """

    return apifunction.ApiFunction.call_(self.name() + '.hypot', self, right)

  @staticmethod
  def identity(size: _arg_types.Integer) -> Array:
    """Returns a 2D identity matrix of the given size.

    Args:
      size: The length of each axis.
    """

    return apifunction.ApiFunction.call_('Array.identity', size)

  def int(self) -> Array:
    """Casts the input value to a signed 32-bit integer."""

    return apifunction.ApiFunction.call_(self.name() + '.int', self)

  def int16(self) -> Array:
    """Casts the input value to a signed 16-bit integer."""

    return apifunction.ApiFunction.call_(self.name() + '.int16', self)

  def int32(self) -> Array:
    """Casts the input value to a signed 32-bit integer."""

    return apifunction.ApiFunction.call_(self.name() + '.int32', self)

  def int64(self) -> Array:
    """Casts the input value to a signed 64-bit integer."""

    return apifunction.ApiFunction.call_(self.name() + '.int64', self)

  def int8(self) -> Array:
    """Casts the input value to a signed 8-bit integer."""

    return apifunction.ApiFunction.call_(self.name() + '.int8', self)

  def lanczos(self) -> Array:
    """Computes the Lanczos approximation of the input."""

    return apifunction.ApiFunction.call_(self.name() + '.lanczos', self)

  def leftShift(self, right: _arg_types.Any) -> Array:
    """On an element-wise basis, calculates the left shift of v1 by v2 bits.

    Args:
      right: How many bits to shift by.

    Returns:
      An ee.Array.
    """

    return apifunction.ApiFunction.call_(
        self.name() + '.leftShift', self, right
    )

  def length(self) -> Array:
    """Returns a 1-D ee.Array containing the length of each dimension."""

    return apifunction.ApiFunction.call_(self.name() + '.length', self)

  def log(self) -> Array:
    """On an element-wise basis, computes the natural logarithm of the input."""

    return apifunction.ApiFunction.call_(self.name() + '.log', self)

  def log10(self) -> Array:
    """On an element-wise basis, computes the base-10 logarithm of the input."""

    return apifunction.ApiFunction.call_(self.name() + '.log10', self)

  def long(self) -> Array:
    """Casts the input value to a signed 64-bit integer."""

    return apifunction.ApiFunction.call_(self.name() + '.long', self)

  def lt(self, right: _arg_types.Any) -> Array:
    """Returns 1 if the first value is less than the second.

    Args:
      right: The values to compare with.

    Returns:
      An ee.Array.
    """

    return apifunction.ApiFunction.call_(self.name() + '.lt', self, right)

  def lte(self, right: _arg_types.Any) -> Array:
    """Returns 1 if the first value is less than or equal to the second.

    Args:
      right: The values to compare with.

    Returns:
      An ee.Array.
    """

    return apifunction.ApiFunction.call_(self.name() + '.lte', self, right)

  def mask(self, mask: _arg_types.Any) -> Array:
    """Creates an array of just the elements that have non-zero mask matches.

    Creates a subarray by slicing out each position in an input array that is
    parallel to a non-zero element of the given mask array.

    Args:
      mask: Mask array.

    Returns:
      An ee.Array.
    """

    return apifunction.ApiFunction.call_(self.name() + '.mask', self, mask)

  def matrixCholeskyDecomposition(self) -> dictionary.Dictionary:
    """Calculates the Cholesky decomposition of a matrix.

    The Cholesky decomposition is a decomposition into the form L * L' where L
    is a lower triangular matrix. The input must be a symmetric
    positive-definite matrix. Returns a dictionary with 1 entry named 'L'.

    Returns:
      An ee.Dictionary.
    """

    return apifunction.ApiFunction.call_(
        self.name() + '.matrixCholeskyDecomposition', self
    )

  def matrixDeterminant(self) -> ee_number.Number:
    """Computes the determinant of the matrix."""

    return apifunction.ApiFunction.call_(
        self.name() + '.matrixDeterminant', self
    )

  def matrixDiagonal(self) -> Array:
    """Computes the diagonal of the matrix in a single column."""

    return apifunction.ApiFunction.call_(self.name() + '.matrixDiagonal', self)

  def matrixFnorm(self) -> ee_number.Number:
    """Computes the Frobenius norm of the matrix."""

    return apifunction.ApiFunction.call_(self.name() + '.matrixFnorm', self)

  def matrixInverse(self) -> Array:
    """Computes the inverse of the matrix."""

    return apifunction.ApiFunction.call_(self.name() + '.matrixInverse', self)

  def matrixLUDecomposition(self) -> dictionary.Dictionary:
    """Calculates the LU matrix decomposition.

    The LU decomposition such that P×input=L×U, where L is lower triangular
    (with unit diagonal terms), U is upper triangular and P is a partial pivot
    permutation matrix.

    The input matrix must be square. Returns a dictionary with entries named
    'L', 'U' and 'P'.

    Returns:
      An ee.Dictionary.
    """

    return apifunction.ApiFunction.call_(
        self.name() + '.matrixLUDecomposition', self
    )

  def matrixMultiply(self, right: _arg_types.Any) -> Array:
    """Returns the matrix multiplication A * B.

    Args:
      right: The right-hand value.

    Returns:
      An ee.Array.
    """

    return apifunction.ApiFunction.call_(
        self.name() + '.matrixMultiply', self, right
    )

  def matrixPseudoInverse(self) -> Array:
    """Computes the Moore-Penrose pseudoinverse of the matrix."""

    return apifunction.ApiFunction.call_(
        self.name() + '.matrixPseudoInverse', self
    )

  def matrixQRDecomposition(self) -> dictionary.Dictionary:
    """Returns the QR decomposition of the input matrix.

    Calculates the QR-decomposition of a matrix into two matrices Q and R such
    that input = QR, where Q is orthogonal, and R is upper triangular.

    Returns a dictionary with entries named 'Q' and 'R'.

    Returns:
      An ee.Dictionary.
    """

    return apifunction.ApiFunction.call_(
        self.name() + '.matrixQRDecomposition', self
    )

  def matrixSingularValueDecomposition(self) -> dictionary.Dictionary:
    """Returns the Singular Value Decomposition of the input matrix.

    Calculates the Singular Value Decomposition of the input matrix into U×S×V',
    such that U and V are orthogonal and S is diagonal.

    Returns:
      An ee.Dictionary with entries named 'U', 'S' and 'V'.
    """

    return apifunction.ApiFunction.call_(
        self.name() + '.matrixSingularValueDecomposition', self
    )

  def matrixSolve(self, right: _arg_types.Any) -> Array:
    """Returns the least-squares solution of the input matrix.

    Solves for x in the matrix equation A * x = B, finding a least-squares
    solution if A is overdetermined.

    Args:
      right: The right-hand value.

    Returns:
      An ee.Array.
    """

    return apifunction.ApiFunction.call_(
        self.name() + '.matrixSolve', self, right
    )

  def matrixToDiag(self) -> Array:
    """Computes a square diagonal matrix from a single column matrix."""

    return apifunction.ApiFunction.call_(self.name() + '.matrixToDiag', self)

  def matrixTrace(self) -> ee_number.Number:
    """Computes the trace of the matrix."""

    return apifunction.ApiFunction.call_(self.name() + '.matrixTrace', self)

  def matrixTranspose(
      self,
      axis1: _arg_types.Integer | None = None,
      axis2: _arg_types.Integer | None = None,
  ) -> Array:
    """Transposes two dimensions of an array.

    Args:
      axis1: First axis to swap. Defaults to 0.
      axis2: Second axis to swap. Defaults to 1.

    Returns:
      An ee.Array.
    """

    return apifunction.ApiFunction.call_(
        self.name() + '.matrixTranspose', self, axis1, axis2
    )

  def max(self, right: _arg_types.Any) -> Array:
    """Selects the maximum of the first and second values.

    Args:
      right: The values to compare with.

    Returns:
      An ee.Array.
    """

    return apifunction.ApiFunction.call_(self.name() + '.max', self, right)

  def min(self, right: _arg_types.Any) -> Array:
    """Selects the minimum of the first and second values.

    Args:
      right: The values to compare with.

    Returns:
      An ee.Array.
    """

    return apifunction.ApiFunction.call_(self.name() + '.min', self, right)

  def mod(self, right: _arg_types.Any) -> Array:
    """Calculates the remainder of the first value divided by the second.

    Args:
      right: The values to divide by.

    Returns:
      An ee.Array.
    """

    return apifunction.ApiFunction.call_(self.name() + '.mod', self, right)

  def multiply(self, right: _arg_types.Any) -> Array:
    """On an element-wise basis, multiplies the first value by the second.

    Args:
      right: The values to multiply by.

    Returns:
      An ee.Array.
    """

    return apifunction.ApiFunction.call_(self.name() + '.multiply', self, right)

  def neq(self, right: _arg_types.Any) -> Array:
    """Returns 1 if the first value is not equal to the second.

    Args:
      right: The right-hand value.

    Returns:
      An ee.Array.
    """

    return apifunction.ApiFunction.call_(self.name() + '.neq', self, right)

  def Not(self) -> Array:
    """Returns 0 if the input is non-zero, and 1 otherwise."""

    return apifunction.ApiFunction.call_(self.name() + '.not', self)

  def Or(self, right: _arg_types.Any) -> Array:
    """On an element-wise basis, returns 1 if either input value is non-zero.

    Args:
      right: The right-hand value.

    Returns:
      An ee.Array.
    """

    return apifunction.ApiFunction.call_(self.name() + '.or', self, right)

  def pad(
      self, lengths: _arg_types.Any, pad: _arg_types.Number | None = None
  ) -> Array:
    """Pad an array to a given length.

    The pad value will be repeatedly appended to the array to extend it to given
    length along each axis. If the array is already as large or larger than a
    given length, it will remain unchanged along that axis.

    Args:
      lengths: A list of new lengths for each axis.
      pad: The value with which to pad the array. Defaults to 0.

    Returns:
      An ee.Array.
    """

    return apifunction.ApiFunction.call_(
        self.name() + '.pad', self, lengths, pad
    )

  def pow(self, right: _arg_types.Any) -> Array:
    """On an element-wise basis, raises the first value to the power of the second.

    Args:
      right: The right-hand value.

    Returns:
      An ee.Array.
    """

    return apifunction.ApiFunction.call_(self.name() + '.pow', self, right)

  def project(self, axes: _arg_types.Any) -> Array:
    """Returns the projected array.

    Projects an array to a lower dimensional space by specifying the axes to
    retain. Dropped axes must be at most length 1.

    Args:
      axes: The axes to project onto. Other axes will be discarded, and must be
        at most length 1.

    Returns:
      An ee.Array.
    """

    return apifunction.ApiFunction.call_(self.name() + '.project', self, axes)

  def reduce(
      self,
      reducer: _arg_types.Reducer,
      axes: _arg_types.Any,
      # pylint: disable-next=invalid-name
      fieldAxis: _arg_types.Integer | None = None,
  ) -> Array:
    """Returns the result of applying the reducer on the array.

    Apply a reducer to an array by collapsing all the input values along each
    specified axis into a single output value computed by the reducer.

    The output always has the same dimensionality as the input, and the
    individual axes are affected as follows:

    - The axes specified in the 'axes' parameter have their length reduced to 1
      (by applying the reducer).
    - If the reducer has multiple inputs or multiple outputs, the axis specified
      in 'fieldAxis' will be used to provide the reducer's inputs and store the
      reducer's outputs.
    - All other axes are unaffected (independent reductions are performed).

    Args:
      reducer: The reducer to apply. Each of its outputs must be a number, not
        an array or other type.
      axes: The list of axes over which to reduce. The output will have a length
        of 1 in all these axes.
      fieldAxis: The axis to use as the reducer's input and output fields. Only
        required if the reducer has multiple inputs or multiple outputs, in
        which case the axis must have length equal to the number of reducer
        inputs, and in the result it will have length equal to the number of
        reducer outputs.

    Returns:
      An ee.Array.
    """

    return apifunction.ApiFunction.call_(
        self.name() + '.reduce', self, reducer, axes, fieldAxis
    )

  def repeat(
      self,
      axis: _arg_types.Integer | None = None,
      copies: _arg_types.Integer | None = None,
  ) -> Array:
    """Repeats the array along the given axis.

    The result will have the shape of the input, except length along the
    repeated axis will be multiplied by the given number of copies.

    Args:
      axis: The axis along which to repeat the array. Defaults to 0.
      copies: The number of copies of this array to concatenate along the given
        axis. Defaults to 2.

    Returns:
      An ee.Array.
    """

    return apifunction.ApiFunction.call_(
        self.name() + '.repeat', self, axis, copies
    )

  def reshape(self, shape: _arg_types.Any) -> Array:
    """Reshapes an array to a new list of dimension lengths.

    Args:
      shape: New shape to which arrays are converted. If one component of the
        shape is the special value -1, the size of that dimension is computed so
        that the total size remains constant. In particular, a shape of [-1]
        flattens into 1-D. At most one component of shape can be -1.

    Returns:
      An ee.Array.
    """

    return apifunction.ApiFunction.call_(self.name() + '.reshape', self, shape)

  def rightShift(self, right: _arg_types.Any) -> Array:
    """Calculates the signed right shift of v1 by v2 bits.

    Args:
      right: The values to shift by.

    Returns:
      An ee.Array.
    """

    return apifunction.ApiFunction.call_(
        self.name() + '.rightShift', self, right
    )

  def round(self) -> Array:
    """On an element-wise basis, computes the integer nearest to the input."""

    return apifunction.ApiFunction.call_(self.name() + '.round', self)

  def short(self) -> Array:
    """On an element-wise basis, casts the input value to a signed 16-bit integer."""

    return apifunction.ApiFunction.call_(self.name() + '.short', self)

  def signum(self) -> Array:
    """Returns -1 or 1 base on the sign of the input or 0 if the input is zero.

    On an element-wise basis, computes the signum function (sign) of the input;
    The return value is 0 if the input is 0, 1 if the input is greater than 0,
    -1 if the input is less than 0.

    Returns:
      An ee.Array.
    """

    return apifunction.ApiFunction.call_(self.name() + '.signum', self)

  def sin(self) -> Array:
    """On an element-wise basis, computes the sine of the input in radians."""

    return apifunction.ApiFunction.call_(self.name() + '.sin', self)

  def sinh(self) -> Array:
    """On an element-wise basis, computes the hyperbolic sine of the input."""

    return apifunction.ApiFunction.call_(self.name() + '.sinh', self)

  def slice(
      self,
      axis: _arg_types.Integer | None = None,
      start: _arg_types.Integer | None = None,
      end: _arg_types.Integer | None = None,
      step: _arg_types.Integer | None = None,
  ) -> Array:
    """Returns the result of slicing the array along the given axis.

    Creates a subarray by slicing out each position along the given axis from
    the 'start' (inclusive) to 'end' (exclusive) by increments of 'step'.

    The result will have as many dimensions as the input, and the same length in
    all directions except the slicing axis, where the length will be the number
    of positions from 'start' to 'end' by 'step' that are in range of the input
    array's length along 'axis'. This means the result can be length 0 along the
    given axis if start=end, or if the start or end values are entirely out of
    range.

    Args:
      axis: The axis to slice on. Defaults to 0.
      start: The coordinate of the first slice (inclusive) along 'axis'.
        Negative numbers are used to position the start of slicing relative to
        the end of the array, where -1 starts at the last position on the axis,
        -2 starts at the next to last position, etc. Defaults to 0.
      end: The coordinate (exclusive) at which to stop taking slices. By default
        this will be the length of the given axis. Negative numbers are used to
        position the end of slicing relative to the end of the array, where -1
        will exclude the last position, -2 will exclude the last two positions,
        etc.
      step: The separation between slices along 'axis'; a slice will be taken at
        each whole multiple of 'step' from 'start' (inclusive) to 'end'
        (exclusive). Must be positive. Defaults to 1.

    Returns:
      An ee.Array.
    """

    return apifunction.ApiFunction.call_(
        self.name() + '.slice', self, axis, start, end, step
    )

  def sort(self, keys: _arg_types.Any | None = None) -> Array:
    """Sorts elements of the array along one axis.

    Args:
      keys: Optional keys to sort by. If not provided, the values are used as
        the keys. The keys can only have multiple elements along one axis, which
        determines the direction to sort in.

    Returns:
      An ee.Array.
    """

    return apifunction.ApiFunction.call_(self.name() + '.sort', self, keys)

  def sqrt(self) -> Array:
    """On an element-wise basis, computes the square root of the input."""

    return apifunction.ApiFunction.call_(self.name() + '.sqrt', self)

  def subtract(self, right: _arg_types.Any) -> Array:
    """On an element-wise basis, subtracts the second value from the first.

    Args:
      right: The values to subtract.

    Returns:
      An ee.Array.
    """

    return apifunction.ApiFunction.call_(self.name() + '.subtract', self, right)

  def tan(self) -> Array:
    """Computes the tangent of the input in radians."""

    return apifunction.ApiFunction.call_(self.name() + '.tan', self)

  def tanh(self) -> Array:
    """Computes the hyperbolic tangent of the input."""

    return apifunction.ApiFunction.call_(self.name() + '.tanh', self)

  def toByte(self) -> Array:
    """Casts the input value to an unsigned 8-bit integer."""

    return apifunction.ApiFunction.call_(self.name() + '.toByte', self)

  def toDouble(self) -> Array:
    """On an element-wise basis, casts the input value to a 64-bit float."""

    return apifunction.ApiFunction.call_(self.name() + '.toDouble', self)

  def toFloat(self) -> Array:
    """On an element-wise basis, casts the input value to a 32-bit float."""

    return apifunction.ApiFunction.call_(self.name() + '.toFloat', self)

  def toInt(self) -> Array:
    """Casts the input value to a signed 32-bit integer."""

    return apifunction.ApiFunction.call_(self.name() + '.toInt', self)

  def toInt16(self) -> Array:
    """Casts the input value to a signed 16-bit integer."""

    return apifunction.ApiFunction.call_(self.name() + '.toInt16', self)

  def toInt32(self) -> Array:
    """Casts the input value to a signed 32-bit integer."""

    return apifunction.ApiFunction.call_(self.name() + '.toInt32', self)

  def toInt64(self) -> Array:
    """Casts the input value to a signed 64-bit integer."""

    return apifunction.ApiFunction.call_(self.name() + '.toInt64', self)

  def toInt8(self) -> Array:
    """Casts the input value to a signed 8-bit integer."""

    return apifunction.ApiFunction.call_(self.name() + '.toInt8', self)

  def toList(self) -> ee_list.List:
    """Turns an Array into a list of lists of numbers."""

    return apifunction.ApiFunction.call_(self.name() + '.toList', self)

  def toLong(self) -> Array:
    """Casts the input value to a signed 64-bit integer."""

    return apifunction.ApiFunction.call_(self.name() + '.toLong', self)

  def toShort(self) -> Array:
    """Casts the input value to a signed 16-bit integer."""

    return apifunction.ApiFunction.call_(self.name() + '.toShort', self)

  def toUint16(self) -> Array:
    """Casts the input value to an unsigned 16-bit integer."""

    return apifunction.ApiFunction.call_(self.name() + '.toUint16', self)

  def toUint32(self) -> Array:
    """Casts the input value to an unsigned 32-bit integer."""

    return apifunction.ApiFunction.call_(self.name() + '.toUint32', self)

  def toUint8(self) -> Array:
    """Casts the input value to an unsigned 8-bit integer."""

    return apifunction.ApiFunction.call_(self.name() + '.toUint8', self)

  def transpose(
      self,
      axis1: _arg_types.Integer | None = None,
      axis2: _arg_types.Integer | None = None,
  ) -> Array:
    """Transposes two dimensions of an array.

    Args:
      axis1: First axis to swap. Defaults to 0.
      axis2: Second axis to swap. Defaults to 1.

    Returns:
      An ee.Array.
    """

    return apifunction.ApiFunction.call_(
        self.name() + '.transpose', self, axis1, axis2
    )

  def trigamma(self) -> Array:
    """Computes the trigamma function of the input."""

    return apifunction.ApiFunction.call_(self.name() + '.trigamma', self)

  def uint16(self) -> Array:
    """Casts the input value to an unsigned 16-bit integer."""

    return apifunction.ApiFunction.call_(self.name() + '.uint16', self)

  def uint32(self) -> Array:
    """Casts the input value to an unsigned 32-bit integer."""

    return apifunction.ApiFunction.call_(self.name() + '.uint32', self)

  def uint8(self) -> Array:
    """Casts the input value to an unsigned 8-bit integer."""

    return apifunction.ApiFunction.call_(self.name() + '.uint8', self)



#!/usr/bin/env python3
"""Tests for the ee.Clusterer module."""

import json
import sys
from typing import Any
import unittest

import unittest
import ee
from ee import apitestcase

_WEKA_COBWEB_SERIALIZED = {
    'result': '0',
    'values': {
        '0': {
            'functionInvocationValue': {
                'functionName': 'Clusterer.wekaCobweb',
                'arguments': {
                    'acuity': {'constantValue': 2},
                    'cutoff': {'constantValue': 0.01},
                    'seed': {'constantValue': 3},
                },
            }
        }
    },
}


def make_expression_graph(
    function_invocation_value: dict[str, Any],
) -> dict[str, Any]:
  return {
      'result': '0',
      'values': {'0': {'functionInvocationValue': function_invocation_value}},
  }


class ClustererTest(apitestcase.ApiTestCase):

  def test_cobweb_args(self):
    clusterer = ee.Clusterer.wekaCobweb(2, 0.01, 3)
    self.assertEqual({'value': 'fakeValue'}, clusterer.getInfo())

    join_func = ee.ApiFunction.lookup('Clusterer.wekaCobweb')
    self.assertEqual(join_func, clusterer.func)
    self.assertFalse(clusterer.isVariable())

    result = json.loads(clusterer.serialize())
    self.assertEqual(_WEKA_COBWEB_SERIALIZED, result)

  def test_cobweb_kwargs(self):
    clusterer = ee.Clusterer.wekaCobweb(acuity=2, cutoff=0.01, seed=3)
    self.assertEqual({'value': 'fakeValue'}, clusterer.getInfo())

    join_func = ee.ApiFunction.lookup('Clusterer.wekaCobweb')
    self.assertEqual(join_func, clusterer.func)
    self.assertFalse(clusterer.isVariable())

    result = json.loads(clusterer.serialize())
    self.assertEqual(_WEKA_COBWEB_SERIALIZED, result)

  def test_cast(self):
    clusterer = ee.Clusterer(
        ee.Clusterer.wekaCobweb(acuity=2, cutoff=0.01, seed=3)
    )
    result = json.loads(clusterer.serialize())
    self.assertEqual(_WEKA_COBWEB_SERIALIZED, result)

  @unittest.skipIf(sys.version_info < (3, 10), 'Unsupported in Python <= 3.9')
  def test_no_args(self):
    message = (
        r'Clusterer\.__init__\(\) missing 1 required positional argument:'
        r' \'clusterer\''
    )
    with self.assertRaisesRegex(TypeError, message):
      ee.Clusterer()  # pytype:disable=missing-parameter

  def test_wrong_type(self):
    message = (
        r'Clusterer can only be used as a cast to Clusterer\. Found <class'
        r' \'int\'>'
    )
    with self.assertRaisesRegex(TypeError, message):
      ee.Clusterer(1234)  # pytype:disable=wrong-arg-types

  def test_schema(self):
    clusterer = ee.Clusterer.wekaCobweb(1.1, 2.2, 3)
    expect = make_expression_graph({
        'functionName': 'Clusterer.schema',
        'arguments': {
            'clusterer': {
                'functionInvocationValue': {
                    'functionName': 'Clusterer.wekaCobweb',
                    'arguments': {
                        'acuity': {'constantValue': 1.1},
                        'cutoff': {'constantValue': 2.2},
                        'seed': {'constantValue': 3},
                    },
                }
            }
        },
    })
    expression = ee.Clusterer(clusterer).schema()
    self.assertIsNotNone(expression)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_train(self):
    clusterer = ee.Clusterer.wekaCobweb(1.1, 2.2, 3)
    features = ee.FeatureCollection('a')
    input_properties = ['b', 'c']
    subsampling = 4.4
    subsampling_seed = 5
    expect = make_expression_graph({
        'arguments': {
            'clusterer': {
                'functionInvocationValue': {
                    'functionName': 'Clusterer.wekaCobweb',
                    'arguments': {
                        'acuity': {'constantValue': 1.1},
                        'cutoff': {'constantValue': 2.2},
                        'seed': {'constantValue': 3},
                    },
                }
            },
            'features': {
                'functionInvocationValue': {
                    'functionName': 'Collection.loadTable',
                    'arguments': {'tableId': {'constantValue': 'a'}},
                }
            },
            'inputProperties': {'constantValue': input_properties},
            'subsampling': {'constantValue': subsampling},
            'subsamplingSeed': {'constantValue': subsampling_seed},
        },
        'functionName': 'Clusterer.train',
    })
    expression = ee.Clusterer(clusterer).train(
        features, input_properties, subsampling, subsampling_seed
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Clusterer(clusterer).train(
        features=features,
        inputProperties=input_properties,
        subsampling=subsampling,
        subsamplingSeed=subsampling_seed,
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_weka_cascade_k_means(self):
    min_clusters = 1
    max_clusters = 2
    restarts = 3
    manual = True
    init = False
    distance_function = 'Manhattan'
    max_iterations = 4
    expect = make_expression_graph({
        'arguments': {
            'minClusters': {'constantValue': min_clusters},
            'maxClusters': {'constantValue': max_clusters},
            'restarts': {'constantValue': restarts},
            'manual': {'constantValue': manual},
            'init': {'constantValue': init},
            'distanceFunction': {'constantValue': distance_function},
            'maxIterations': {'constantValue': max_iterations},
        },
        'functionName': 'Clusterer.wekaCascadeKMeans',
    })
    expression = ee.Clusterer.wekaCascadeKMeans(
        min_clusters,
        max_clusters,
        restarts,
        manual,
        init,
        distance_function,
        max_iterations,
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Clusterer.wekaCascadeKMeans(
        minClusters=min_clusters,
        maxClusters=max_clusters,
        restarts=restarts,
        manual=manual,
        init=init,
        distanceFunction=distance_function,
        maxIterations=max_iterations,
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_weka_cobweb(self):
    acuity = 1.0
    cutoff = 2.0
    seed = 3
    expect = make_expression_graph({
        'arguments': {
            'acuity': {'constantValue': acuity},
            'cutoff': {'constantValue': cutoff},
            'seed': {'constantValue': seed},
        },
        'functionName': 'Clusterer.wekaCobweb',
    })
    expression = ee.Clusterer.wekaCobweb(acuity, cutoff, seed)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Clusterer.wekaCobweb(
        acuity=acuity, cutoff=cutoff, seed=seed
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_weka_k_means(self):
    n_clusters = 1
    init = 2
    canopies = True
    max_candidates = 3
    periodic_pruning = 4
    min_density = 5
    t1 = 6.1
    t2 = 7.1
    distance_function = 'Euclidean'
    max_iterations = 8
    preserve_order = False
    fast = True
    seed = 9
    expect = make_expression_graph({
        'arguments': {
            'nClusters': {'constantValue': n_clusters},
            'init': {'constantValue': init},
            'canopies': {'constantValue': canopies},
            'maxCandidates': {'constantValue': max_candidates},
            'periodicPruning': {'constantValue': periodic_pruning},
            'minDensity': {'constantValue': min_density},
            't1': {'constantValue': t1},
            't2': {'constantValue': t2},
            'distanceFunction': {'constantValue': distance_function},
            'maxIterations': {'constantValue': max_iterations},
            'preserveOrder': {'constantValue': preserve_order},
            'fast': {'constantValue': fast},
            'seed': {'constantValue': seed},
        },
        'functionName': 'Clusterer.wekaKMeans',
    })
    expression = ee.Clusterer.wekaKMeans(
        n_clusters,
        init,
        canopies,
        max_candidates,
        periodic_pruning,
        min_density,
        t1,
        t2,
        distance_function,
        max_iterations,
        preserve_order,
        fast,
        seed,
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Clusterer.wekaKMeans(
        nClusters=n_clusters,
        init=init,
        canopies=canopies,
        maxCandidates=max_candidates,
        periodicPruning=periodic_pruning,
        minDensity=min_density,
        t1=t1,
        t2=t2,
        distanceFunction=distance_function,
        maxIterations=max_iterations,
        preserveOrder=preserve_order,
        fast=fast,
        seed=seed,
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_weka_lvq(self):
    num_clusters = 1
    learning_rate = 2.2
    epochs = 3
    normalize_input = True
    expect = make_expression_graph({
        'arguments': {
            'numClusters': {'constantValue': num_clusters},
            'learningRate': {'constantValue': learning_rate},
            'epochs': {'constantValue': epochs},
            'normalizeInput': {'constantValue': normalize_input},
        },
        'functionName': 'Clusterer.wekaLVQ',
    })
    expression = ee.Clusterer.wekaLVQ(
        num_clusters, learning_rate, epochs, normalize_input
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Clusterer.wekaLVQ(
        numClusters=num_clusters,
        learningRate=learning_rate,
        epochs=epochs,
        normalizeInput=normalize_input,
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_weka_x_means(self):
    min_clusters = 1
    max_clusters = 2
    max_iterations = 3
    max_k_means = 4
    max_for_children = 5
    use_k_d = True
    cutoff_factor = 6.1
    distance_function = 'Chebyshev'
    seed = 7
    expect = make_expression_graph({
        'arguments': {
            'minClusters': {'constantValue': min_clusters},
            'maxClusters': {'constantValue': max_clusters},
            'maxIterations': {'constantValue': max_iterations},
            'maxKMeans': {'constantValue': max_k_means},
            'maxForChildren': {'constantValue': max_for_children},
            'useKD': {'constantValue': use_k_d},
            'cutoffFactor': {'constantValue': cutoff_factor},
            'distanceFunction': {'constantValue': distance_function},
            'seed': {'constantValue': seed},
        },
        'functionName': 'Clusterer.wekaXMeans',
    })
    expression = ee.Clusterer.wekaXMeans(
        min_clusters,
        max_clusters,
        max_iterations,
        max_k_means,
        max_for_children,
        use_k_d,
        cutoff_factor,
        distance_function,
        seed,
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Clusterer.wekaXMeans(
        minClusters=min_clusters,
        maxClusters=max_clusters,
        maxIterations=max_iterations,
        maxKMeans=max_k_means,
        maxForChildren=max_for_children,
        useKD=use_k_d,
        cutoffFactor=cutoff_factor,
        distanceFunction=distance_function,
        seed=seed,
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)


if __name__ == '__main__':
  unittest.main()



"""A wrapper for Joins."""

from __future__ import annotations

from ee import _arg_types
from ee import apifunction
from ee import computedobject
from ee import featurecollection


class Join(computedobject.ComputedObject):
  """An object to represent an Earth Engine Join.

  Example:
    fc1 = ee.FeatureCollection([
        ee.Feature(None, {'label': 1}),
        ee.Feature(None, {'label': 2}),
    ])
    fc2 = ee.FeatureCollection([
        ee.Feature(None, {'label': 1}),
        ee.Feature(None, {'label': 3}),
    ])
    a_filter = ee.Filter.equals(leftField='label', rightField='label')
    join = ee.Join.simple()
    joined = join.apply(fc1, fc2, a_filter)
  """

  _initialized: bool = False

  def __init__(
      self,
      join: computedobject.ComputedObject,
  ):
    """Creates a Join wrapper.

    Args:
      join: A join to cast.
    """
    self.initialize()

    if isinstance(join, computedobject.ComputedObject):
      # There is no server-side constructor for ee.Join. Pass the object as-is
      # to the server in case it is intended to be a Join cast.
      super().__init__(join.func, join.args, join.varName)
      return

    raise TypeError(
        f'Join can only be used as a cast to Join. Found {type(join)}.')

  @classmethod
  def initialize(cls) -> None:
    """Imports API functions to this class."""
    if not cls._initialized:
      apifunction.ApiFunction.importApi(cls, cls.name(), cls.name())
      cls._initialized = True

  @classmethod
  def reset(cls) -> None:
    """Removes imported API functions from this class."""
    apifunction.ApiFunction.clearApi(cls)
    cls._initialized = False

  @staticmethod
  def name() -> str:
    return 'Join'

  def apply(
      self,
      primary: _arg_types.FeatureCollection,
      secondary: _arg_types.FeatureCollection,
      condition: _arg_types.Filter,
  ) -> featurecollection.FeatureCollection:
    """Joins two collections.

    Args:
      primary: The primary collection.
      secondary: The secondary collection.
      condition: The join condition used to select the matches from the two
        collections.

    Returns:
      An ee.FeatureCollection.
    """

    return apifunction.ApiFunction.call_(
        self.name() + '.apply', self, primary, secondary, condition
    )

  @staticmethod
  def inner(
      # pylint: disable=invalid-name
      primaryKey: _arg_types.String | None = None,
      secondaryKey: _arg_types.String | None = None,
      measureKey: _arg_types.String | None = None,
      # pylint: enable=invalid-name
  ) -> Join:
    """Returns a join with matching pairs of elements.

    Returns a join that pairs elements from the primary collection with matching
    elements from the secondary collection.

    Each result has a 'primary' property that contains the element from the
    primary collection, and a 'secondary' property containing the matching
    element from the secondary collection. If measureKey is specified, the join
    measure is also attached to the object as a property.

    Args:
      primaryKey: The property name used to save the primary match.
      secondaryKey: The property name used to save the secondary match.
      measureKey: An optional property name used to save the measure of the join
        condition.
    """

    return apifunction.ApiFunction.call_(
        'Join.inner', primaryKey, secondaryKey, measureKey
    )

  @staticmethod
  def inverted() -> Join:
    """Returns a join producing only elements not in the secondary collection.

    Returns a join that produces the elements of the primary collection that
    match no elements of the secondary collection. No properties are added to
    the results.
    """

    return apifunction.ApiFunction.call_('Join.inverted')

  @staticmethod
  def saveAll(
      # pylint: disable-next=invalid-name
      matchesKey: _arg_types.String | None = None,
      ordering: _arg_types.String | None = None,
      ascending: _arg_types.Bool | None = None,
      # pylint: disable-next=invalid-name
      measureKey: _arg_types.String | None = None,
      outer: _arg_types.Bool | None = None,
  ) -> Join:
    """Returns a join that returns all pairs of elements.

    Returns a join that pairs each element from the first collection with a
    group of matching elements from the second collection.

    The list of matches is added to each result as an additional property. If
    measureKey is specified, each match has the value of its join measure
    attached. Join measures are produced when withinDistance or maxDifference
    filters are used as the join condition.

    Args:
      matchesKey: The property name used to save the matches list.
      ordering: The property on which to sort the matches list.
      ascending: Whether the ordering is ascending.
      measureKey: An optional property name used to save the measure of the join
        condition on each match.
      outer: If true, primary rows without matches will be included in the
        result.
    """

    return apifunction.ApiFunction.call_(
        'Join.saveAll', matchesKey, ordering, ascending, measureKey, outer
    )

  @staticmethod
  def saveBest(
      matchKey: _arg_types.String,  # pylint: disable=invalid-name
      measureKey: _arg_types.String,  # pylint: disable=invalid-name
      outer: _arg_types.Bool | None = None,
  ) -> Join:
    """Returns a join that returns the best match pairs.

    Returns a join that pairs each element from the first collection with a
    matching element from the second collection. The match with the best join
    measure is added to each result as an additional property. Join measures are
    produced when withinDistance or maxDifference filters are used as the join
    condition.

    Args:
      matchKey: The key used to save the match.
      measureKey: The key used to save the measure of the join condition on the
        match.
      outer: If true, primary rows without matches will be included in the
        result.
    """

    return apifunction.ApiFunction.call_(
        'Join.saveBest', matchKey, measureKey, outer
    )

  @staticmethod
  def saveFirst(
      matchKey: _arg_types.String,  # pylint: disable=invalid-name
      ordering: _arg_types.String | None = None,
      ascending: _arg_types.Bool | None = None,
      # pylint: disable-next=invalid-name
      measureKey: _arg_types.String | None = None,
      outer: _arg_types.Bool | None = None,
  ) -> Join:
    """Returns a join that returns the first match pairs.

    Returns a join that pairs each element from the first collection with a
    matching element from the second collection. The first match is added to the
    result as an additional property.

    Args:
      matchKey: The property name used to save the match.
      ordering: The property on which to sort the matches before selecting the
        first.
      ascending: Whether the ordering is ascending.
      measureKey: An optional property name used to save the measure of the join
        condition on the match.
      outer: If true, primary rows without matches will be included in the
        result.
    """

    return apifunction.ApiFunction.call_(
        'Join.saveFirst', matchKey, ordering, ascending, measureKey, outer
    )

  @staticmethod
  def simple() -> Join:
    """Returns a join that returns match pairs.

    Returns a join that produces the elements of the primary collection that
    match any element of the secondary collection. No properties are added to
    the results.
    """

    return apifunction.ApiFunction.call_('Join.simple')



"""Representation of an Earth Engine FeatureCollection."""

# Using lowercase function naming to match the JavaScript names.
# pylint: disable=g-bad-name

from __future__ import annotations

from typing import Any

from ee import _arg_types
from ee import _utils
from ee import apifunction
from ee import collection
from ee import computedobject
from ee import data
from ee import deprecation
from ee import ee_exception
from ee import ee_list
from ee import ee_types
from ee import feature
from ee import geometry
from ee import image


class FeatureCollection(collection.Collection[feature.Feature]):
  """A representation of a FeatureCollection."""

  _initialized = False

  # Tell pytype to not complain about dynamic attributes.
  _HAS_DYNAMIC_ATTRIBUTES = True

  @_utils.accept_opt_prefix('opt_column')
  @deprecation.WarnForDeprecatedAsset('args')
  def __init__(
      self,
      args: None | (
          dict[str, Any] |
          list[Any] |
          str |
          feature.Feature |
          geometry.Geometry |
          computedobject.ComputedObject
      ),
      column: Any | None = None,
  ):
    """Constructs a collection features.

    Args:
      args: constructor argument. One of:
        1) A string - assumed to be the name of a collection.
        2) A geometry.
        3) A feature.
        4) An array of features.
        5) A GeoJSON FeatureCollection.
        6) A computed object - reinterpreted as a collection.
      column: The name of the geometry column to use. Only useful with the
        string constructor.

    Raises:
      EEException: if passed something other than the above.
    """
    self.initialize()

    # Wrap geometries with features.
    if isinstance(args, geometry.Geometry):
      args = feature.Feature(args)

    # Wrap single features in an array.
    if isinstance(args, feature.Feature):
      args = [args]

    if ee_types.isString(args):
      # An ID.
      actual_args = {'tableId': args}
      if column:
        actual_args['geometryColumn'] = column
      super().__init__(
          apifunction.ApiFunction.lookup('Collection.loadTable'), actual_args)
    elif isinstance(args, (list, tuple)):
      # A list of features.
      super().__init__(
          apifunction.ApiFunction.lookup('Collection'), {
              'features': [feature.Feature(i) for i in args]
          })
    elif isinstance(args, ee_list.List):
      # A computed list of features.
      super().__init__(
          apifunction.ApiFunction.lookup('Collection'), {
              'features': args
          })
    elif isinstance(args, dict) and args.get('type') == self.name():
      # A GeoJSON FeatureCollection
      super().__init__(
          apifunction.ApiFunction.lookup('Collection'),
          {'features': [feature.Feature(i) for i in args.get('features', [])]})
    elif isinstance(args, computedobject.ComputedObject):
      # A custom object to reinterpret as a FeatureCollection.
      super().__init__(args.func, args.args, args.varName)
    else:
      raise ee_exception.EEException(
          'Unrecognized argument type to convert to a FeatureCollection: %s' %
          args)

  @classmethod
  def initialize(cls) -> None:
    """Imports API functions to this class."""
    if not cls._initialized:
      super().initialize()
      apifunction.ApiFunction.importApi(cls, cls.name(), cls.name())
      cls._initialized = True

  @classmethod
  def reset(cls) -> None:
    """Removes imported API functions from this class."""
    apifunction.ApiFunction.clearApi(cls)
    cls._initialized = False

  def getMapId(
      self, vis_params: dict[str, Any] | None = None
  ) -> dict[str, Any]:
    """Fetch and return a map id and token, suitable for use in a Map overlay.

    Args:
      vis_params: The visualization parameters. Currently only one parameter,
          'color', containing a hex RGB color string is allowed.

    Returns:
      A map ID dictionary as described in ee.data.getMapId, including an
      additional 'image' field containing Collection.draw image wrapping a
      FeatureCollection containing this feature.
    """
    painted = apifunction.ApiFunction.apply_('Collection.draw', {
        'collection': self,
        'color': (vis_params or {}).get('color', '000000')
    })
    return painted.getMapId({})

  def getDownloadURL(
      self,
      filetype: str | None = None,
      selectors: Any | None = None,
      filename: str | None = None,
  ) -> str:
    """Gets a download URL.

    When the URL is accessed, the FeatureCollection is downloaded in one of
    several formats.

    Args:
      filetype: The format of download, one of: "csv", "json", "geojson", "kml",
          "kmz" ("json" outputs GeoJSON). If unspecified, defaults to "csv".
      selectors: Feature property names used to select the attributes to be
          downloaded. If unspecified, all properties are included.
      filename: Name of the file to be downloaded; extension is appended by
          default. If unspecified, defaults to "table".

    Returns:
      A URL to download this FeatureCollection.
    """
    request = {}
    request['table'] = self
    if filetype is not None:
      request['format'] = filetype.upper()
    if filename is not None:
      request['filename'] = filename
    if selectors is not None:
      if isinstance(selectors, (list, tuple)):
        selectors = ','.join(selectors)
      request['selectors'] = selectors
    return data.makeTableDownloadUrl(data.getTableDownloadId(request))

  # Deprecated spelling to match the JS library.
  getDownloadUrl = deprecation.Deprecated('Use getDownloadURL().')(
      getDownloadURL)

  # TODO(user): How to handle type annotations for
  #  feature_collection.select('a', 'b', 'c')?
  # pylint: disable-next=keyword-arg-before-vararg
  def select(
      self,
      propertySelectors: Any,
      newProperties: Any | None = None,
      retainGeometry: bool | str = True,
      *args,
  ) -> FeatureCollection:
    """Select properties from each feature in a collection.

    Args:
      propertySelectors: An array of names or regexes specifying the properties
          to select.
      newProperties: An array of strings specifying the new names for the
          selected properties. If supplied, the length must match the number
          of properties selected.
      retainGeometry: A boolean. When false, the result will have no geometry.
      *args: Selector elements as varargs.

    Returns:
      The feature collection with selected properties.
    """
    if len(args) or ee_types.isString(propertySelectors):
      args = list(args)
      if not isinstance(retainGeometry, bool):
        args.insert(0, retainGeometry)
      if newProperties is not None:
        args.insert(0, newProperties)
      args.insert(0, propertySelectors)
      return self.map(lambda feat: feat.select(args, None, True))
    else:
      return self.map(
          # pylint: disable-next=g-long-lambda
          lambda feat: feat.select(
              propertySelectors, newProperties, retainGeometry
          )
      )

  @staticmethod
  def name() -> str:
    return 'FeatureCollection'

  @staticmethod
  def elementType() -> type[feature.Feature]:
    return feature.Feature

  def classify(
      self,
      classifier: _arg_types.Classifier,
      # pylint: disable-next=invalid-name
      outputName: _arg_types.String | None = None,
  ) -> FeatureCollection:
    """Returns the result of classifying each feature in a collection.

    Args:
      classifier: The classifier to use.
      outputName: The name of the output property to be added. This argument is
        ignored if the classifier has more than one output.

    Returns:
      An ee.FeatureCollection.
    """

    return apifunction.ApiFunction.call_(
        self.name() + '.classify', self, classifier, outputName
    )

  def cluster(
      self,
      clusterer: _arg_types.Clusterer,
      # pylint: disable-next=invalid-name
      outputName: _arg_types.String | None = None,
  ) -> FeatureCollection:
    """Returns the results of clustering each feature in a collection.

    Clusters each feature in a collection, adding a new column to each feature
    containing the cluster number to which it has been assigned.

    Args:
      clusterer: The clusterer to use.
      outputName: The name of the output property to be added.
    """

    return apifunction.ApiFunction.call_(
        self.name() + '.cluster', self, clusterer, outputName
    )

  def inverseDistance(
      self,
      range: _arg_types.Number,  # pylint: disable=redefined-builtin
      propertyName: _arg_types.String,  # pylint: disable=invalid-name
      mean: _arg_types.Number,
      stdDev: _arg_types.Number,  # pylint: disable=invalid-name
      gamma: _arg_types.Number | None = None,
      reducer: _arg_types.Reducer | None = None,
  ) -> image.Image:
    """Returns an inverse-distance weighted estimate of the value at each pixel.

    Args:
      range: Size of the interpolation window (in meters).
      propertyName: Name of the numeric property to be estimated.
      mean: Global expected mean.
      stdDev: Global standard deviation.
      gamma: Determines how quickly the estimates tend towards the global mean.
      reducer: Reducer used to collapse the 'propertyName' value of overlapping
        points into a single value.
    """

    return apifunction.ApiFunction.call_(
        self.name() + '.inverseDistance',
        self,
        range,
        propertyName,
        mean,
        stdDev,
        gamma,
        reducer,
    )

  def kriging(
      self,
      propertyName: _arg_types.String,  # pylint: disable=invalid-name
      shape: _arg_types.String,
      range: _arg_types.Number,  # pylint: disable=redefined-builtin
      sill: _arg_types.Number,
      nugget: _arg_types.Number,
      # pylint: disable-next=invalid-name
      maxDistance: _arg_types.Number | None = None,
      reducer: _arg_types.Reducer | None = None,
  ) -> image.Image:
    """Returns the results of sampling a Kriging estimator at each pixel.

    Args:
      propertyName: Property to be estimated (must be numeric).
      shape: Semivariogram shape (one of {exponential, gaussian, spherical}).
      range: Semivariogram range, in meters.
      sill: Semivariogram sill.
      nugget: Semivariogram nugget.
      maxDistance: Radius which determines which features are included in each
        pixel's computation, in meters. Defaults to the semivariogram's range.
      reducer: Reducer used to collapse the 'propertyName' value of overlapping
        points into a single value.
    """

    return apifunction.ApiFunction.call_(
        self.name() + '.kriging',
        self,
        propertyName,
        shape,
        range,
        sill,
        nugget,
        maxDistance,
        reducer,
    )

  @staticmethod
  def loadBigQueryTable(
      table: _arg_types.String,
      geometryColumn: _arg_types.String | None = None,
  ) -> FeatureCollection:
    """Returns a FeatureCollection containing data read from a BigQuery table.

    Args:
      table: Path to BigQuery table in a project.dataset.table format.
      geometryColumn: The name of the column to use as the main feature
        geometry. If not specified, all features will have null geometry.
    """

    return apifunction.ApiFunction.call_(
        'FeatureCollection.loadBigQueryTable',
        table,
        geometryColumn,
    )

  def makeArray(
      self,
      properties: _arg_types.List,
      name: _arg_types.String | None = None,
  ) -> FeatureCollection:
    """Returns a collection with a 1-D Array property for each feature.

    Add a 1-D Array to each feature in a collection by combining a list of
    properties for each feature into a 1-D Array.

    All of the properties must be numeric values. If a feature doesn't contain
    all of the named properties, or any of them aren't numeric, the feature will
    be dropped from the resulting collection.

    Args:
      properties: The properties to select.
      name: The name of the new array property.
    """

    return apifunction.ApiFunction.call_(
        self.name() + '.makeArray', self, properties, name
    )

  @staticmethod
  def randomPoints(
      region: geometry.Geometry,
      points: _arg_types.Integer | None = None,
      seed: _arg_types.Integer | None = None,
      # pylint: disable-next=invalid-name
      maxError: _arg_types.ErrorMargin | None = None,
  ) -> FeatureCollection:
    """Returns a collection of random points.

    Generates points that are uniformly random in the given geometry. If the
    geometry is two-dimensional (polygon or multi-polygon) then the returned
    points are uniformly distributed on the given region of the sphere. If the
    geometry is one-dimensional (linestrings), the returned points are
    interpolated uniformly along the geometry's edges. If the geometry has
    dimension zero (points), the returned points are sampled uniformly from the
    input points. If a multi-geometry of mixed dimension is given, points are
    sampled from the component geometries with the highest dimension.

    Args:
      region: The region to generate points for.
      points: The number of points to generate.
      seed: A seed for the random number generator.
      maxError: The maximum amount of error tolerated when performing any
        necessary reprojection.
    """

    return apifunction.ApiFunction.call_(
        'FeatureCollection.randomPoints', region, points, seed, maxError
    )

  @staticmethod
  def runBigQuery(
      query: _arg_types.String,
      geometryColumn: _arg_types.String | None = None,
      maxBytesBilled: _arg_types.Integer | None = int(1e11),
  ) -> FeatureCollection:
    """Returns a FeatureCollection containing result of a BigQuery query.

    Args:
      query: GoogleSQL query to perform on the BigQuery resources.
      geometryColumn: The name of the column to use as the main feature
        geometry. If not specified, all features will have null geometry.
      maxBytesBilled: Maximum number of bytes billed while processing the query.
        Any BigQuery job that exceeds this limit will fail and won't be billed.
        Defaults to 100GB.
    """

    return apifunction.ApiFunction.call_(
        'FeatureCollection.runBigQuery',
        query,
        geometryColumn,
        maxBytesBilled
    )



"""A serializer that encodes EE object trees as JSON DAGs."""

import collections
import datetime
import hashlib
import json
from typing import Any

from ee import _cloud_api_utils
from ee import _utils
from ee import ee_exception
from ee import encodable

# The datetime for the beginning of the Unix epoch.
_EPOCH_DATETIME = datetime.datetime.fromtimestamp(0, datetime.timezone.utc)

# Don't generate very deep expressions, as the backend rejects them.
# The backend's limit is 100, and we want to stay well away from that
# as a few extra levels of wrapping are always added.
_DEPTH_LIMIT = 50


# pylint: disable-next=g-bad-name
def DatetimeToMicroseconds(date: datetime.datetime) -> int:
  """Convert a datetime to a timestamp, microseconds since the epoch."""
  if date.tzinfo is None:
    # Assume that the time is in utc.
    date = date.replace(tzinfo=datetime.timezone.utc)
  td = (date - _EPOCH_DATETIME)
  return td.microseconds + (td.seconds + td.days * 24 * 3600) * 1000000


class Serializer:
  """A serializer for EE object trees."""
  unbound_name: str | None

  # Whether the encoding should factor out shared subtrees.
  _is_compound: bool
  _for_cloud_api: bool
  # A list of shared subtrees as [name, value] pairs.
  _scope: list[str]
  # A lookup table from object hash to subtree names as stored in self._scope
  _encoded: dict[Any, Any]
  # A lookup table from object ID as retrieved by id() to md5 hash values.
  _hashcache: dict[Any, Any]

  def __init__(
      self,
      is_compound: bool = True,
      for_cloud_api: bool = False,
      unbound_name: str | None = None,
  ):
    """Constructs a serializer.

    Args:
      is_compound: Whether the encoding should factor out shared subtrees.
      for_cloud_api: Whether the encoding should be done for the Cloud API or
        the legacy API.
      unbound_name: Provides a name for unbound variables in objects.
    """
    self.unbound_name = unbound_name

    self._is_compound = is_compound
    self._for_cloud_api = for_cloud_api
    self._scope = []
    self._encoded = {}
    self._hashcache = {}

  def _encode(self, obj: Any) -> Any:
    """Encodes a top level object to be executed server-side.

    Args:
      obj: The object to encode.

    Returns:
      An encoded object ready for JSON serialization.
    """
    if self._for_cloud_api:
      return self._encode_for_cloud_api(obj)
    value = self._encode_value(obj)
    if self._is_compound:
      if (isinstance(value, dict) and value['type'] == 'ValueRef' and
          len(self._scope) == 1):
        # Just one value. No need for complex structure.
        value = self._scope[0][1]
      else:
        # Wrap the scopes and final value with a CompoundValue.
        value = {'type': 'CompoundValue', 'scope': self._scope, 'value': value}
      # Clear state in case of future encoding.
      self._scope = []
      self._encoded = {}
      self._hashcache = {}
    return value

  def _encode_for_cloud_api(self, obj: Any) -> Any:
    """Encodes an object as an Expression or quasi-Expression."""
    value = self._encode_cloud_object(obj)
    if self._is_compound:
      # Wrap the scopes and final value into an Expression.
      value = _ExpressionOptimizer(value, self._scope).optimize()
      # Clear state in case of future encoding.
      self._scope = []
      self._encoded = {}
      self._hashcache = {}
    else:
      value = _ExpressionOptimizer(value).optimize()
    return value

  def _encode_value(self, obj: Any) -> Any:
    """Encodes a subtree as a Value in the EE API v2 (DAG) format.

    If _is_compound is True, this will fill the _scope and _encoded properties.

    Args:
      obj: The object to encode.

    Returns:
      An encoded object.
    """
    obj_id = id(obj)
    hashval = self._hashcache.get(obj_id)
    encoded = self._encoded.get(hashval, None)
    if self._is_compound and encoded:
      # Already encoded objects are encoded as ValueRefs and returned directly.
      return {'type': 'ValueRef', 'value': encoded}
    elif obj is None or isinstance(obj, (bool, float, int, str)):
      # Primitives are encoded as is and not saved in the scope.
      return obj
    elif isinstance(obj, datetime.datetime):
      # A raw date slipped through. Wrap it. Calling ee.Date from here would
      # cause a circular dependency, so we encode it manually.
      return {
          'type': 'Invocation',
          'functionName': 'Date',
          'arguments': {
              'value': DatetimeToMicroseconds(obj) / 1e3
          }
      }
    elif isinstance(obj, encodable.Encodable):
      # Some objects know how to encode themselves.
      result = obj.encode(self._encode_value)
      if (not isinstance(result, (list, tuple)) and
          (not isinstance(result, (dict)) or result['type'] == 'ArgumentRef')):
        # Optimization: simple enough that adding it to the scope is probably
        # not worth it.
        return result
    elif isinstance(obj, encodable.EncodableFunction):
      result = obj.encode_invocation(self._encode_value)
      if (not isinstance(result, (list, tuple)) and
          (not isinstance(result, (dict)) or result['type'] == 'ArgumentRef')):
        # Optimization: simple enough that adding it to the scope is probably
        # not worth it.
        return result
    elif isinstance(obj, (list, tuple)):
      # Lists are encoded recursively.
      result = [self._encode_value(i) for i in obj]
    elif isinstance(obj, dict):
      # Dictionary are encoded recursively and wrapped in a type specifier.
      result = {
          'type':
              'Dictionary',
          'value':
              {key: self._encode_value(value)
                    for key, value in obj.items()}
      }
    else:
      raise ee_exception.EEException('Cannot encode object: %s' % obj)

    if self._is_compound:
      # Save the new object and return a ValueRef.
      hashval = hashlib.md5(json.dumps(result).encode()).digest()
      self._hashcache[obj_id] = hashval
      name = self._encoded.get(hashval, None)
      if not name:
        name = str(len(self._scope))
        self._scope.append((name, result))
        self._encoded[hashval] = name
      return {'type': 'ValueRef', 'value': name}
    else:
      return result

  def _encode_cloud_object(self, obj: Any) -> Any:
    """Encodes an object using the Cloud API Expression form.

    If _is_compound is True, this will fill the _scope and _encoded properties.

    Args:
      obj: The object to encode.

    Returns:
      If _is_compound is True, a string that is the key under which the
        encoded object is stored in _scope.
      If _is_compound is False, the encoded object as a single quasi-Expression.
    """
    obj_id = id(obj)
    hashval = self._hashcache.get(obj_id)
    reference = self._encoded.get(hashval, None)
    if reference:
      return reference
    elif obj is None or isinstance(obj, (bool, str)):
      result = {'constantValue': obj}
    elif isinstance(obj, (float, int)):
      result = _cloud_api_utils.encode_number_as_cloud_value(obj)
    elif isinstance(obj, datetime.datetime):
      # A raw date slipped through. Wrap it. Calling ee.Date from here would
      # cause a circular dependency, so we encode it manually.
      result = {
          'functionInvocationValue': {
              'functionName': 'Date',
              'arguments': {
                  'value': {
                      'constantValue': DatetimeToMicroseconds(obj) / 1e3
                  }
              }
          }
      }
    elif isinstance(obj, encodable.Encodable):
      # Some objects know how to encode themselves.
      result = obj.encode_cloud_value(self._encode_cloud_object)
    elif isinstance(obj, (list, tuple)):
      # Lists are encoded recursively.
      if self._is_compound:
        result = {
            'arrayValue': {
                'values': [{
                    'valueReference': self._encode_cloud_object(i)
                } for i in obj]
            }
        }
      else:
        result = {
            'arrayValue': {
                'values': [self._encode_cloud_object(i) for i in obj]
            }
        }
    elif isinstance(obj, dict):
      # Dictionary are encoded recursively and wrapped in a type specifier.
      # We iterate through the entries in a deterministic order, not because it
      # affects the order of the entries in the output result, but because it
      # affects the names that they are assigned in _scope; without the
      # ordering, the encoding process may produce one of multiple different
      # (albeit equivalent) representations.
      if self._is_compound:
        result = {
            'dictionaryValue': {
                'values': {
                    key: {
                        'valueReference': self._encode_cloud_object(obj[key])
                    } for key in sorted(obj)
                }
            }
        }
      else:
        result = {
            'dictionaryValue': {
                'values': {
                    key: self._encode_cloud_object(obj[key])
                    for key in sorted(obj)
                }
            }
        }
    else:
      raise ee_exception.EEException('Cannot encode object: %s' % obj)

    if self._is_compound:
      # Save the new object and return a ValueRef.
      hashval = hashlib.md5(json.dumps(result).encode()).digest()
      self._hashcache[obj_id] = hashval
      name = self._encoded.get(hashval, None)
      if not name:
        name = str(len(self._scope))
        self._scope.append((name, result))
        self._encoded[hashval] = name
      return name
    else:
      return result


def encode(
    obj: Any,
    is_compound: bool = True,
    for_cloud_api: bool = True,
    unbound_name: str | None = None,
) -> Any:
  """Serialize an object to a JSON-compatible structure for API calls.

  Args:
    obj: The object to serialize.
    is_compound: Whether the encoding should factor out shared subtrees.
    for_cloud_api: Whether the encoding should be done for the Cloud API or the
      legacy API.
    unbound_name: Provides a name for unbound variables in objects. Unbound
      variables are otherwise disallowed. See the Count Functions usage in
      customfunction.py.

  Returns:
    A JSON-compatible structure representing the input.
  """
  serializer = Serializer(
      is_compound, for_cloud_api=for_cloud_api, unbound_name=unbound_name)
  return serializer._encode(obj)  # pylint: disable=protected-access


@_utils.accept_opt_prefix('opt_pretty')
# pylint: disable-next=g-bad-name
def toJSON(obj, pretty: bool = False, for_cloud_api: bool = True) -> Any:
  """Serialize an object to a JSON string appropriate for API calls.

  Args:
    obj: The object to serialize.
    pretty: True to pretty-print the object.
    for_cloud_api: Whether the encoding should be done for the Cloud API or the
      legacy API.

  Returns:
    A JSON string representing the input.
  """
  serializer = Serializer(not pretty, for_cloud_api=for_cloud_api)
  encoded = serializer._encode(obj)  # pylint: disable=protected-access
  return json.dumps(encoded, indent=2 if pretty else None)


# pylint: disable-next=g-bad-name
def toReadableJSON(obj: Any, for_cloud_api: bool = True) -> Any:
  """Convert an object to readable JSON."""
  return toJSON(obj, True, for_cloud_api=for_cloud_api)


class _ExpressionOptimizer:
  """Optimises the representation of an Expression.

  The Expressions generated by recursive encoding can be inefficiently
  represented. This class helps improve the representation.

  The initial representation is intentionally simple, as it makes the encoding
  logic simple. Constants end up as individual ValueNodes, though the Expression
  format itself allows complex constants (nested arrays and/or dicts containing
  constant values). There are also often places where references to ValueNodes
  can be replaced by direct inclusion of those ValueNodes.

  This operates in two modes:
  - It can be passed an Expression as a dict of named ValueNodes, and the name
    that represents the final result. In this case, it returns the optimised
    Expression in the same form. This is the "compound" mode.
  - It can be passed a quasi-Expression as a single object. In this case, it
    returns the optimised quasi-Expression in the same form. This is the
    "non-compound" mode. A "quasi-Expression" is essentially an Expression DAG
    that's been expanded to a tree by replacing references with the actual thing
    being referenced. This means that if the same entity is referenced more than
    once, it will be duplicated in the tree.

  The rules that the optimiser follows are straightforward:
  - If a value is referred to only once, lift it into the place that references
    it.
  - If a value is a numeric or boolean constant, lift it into all the places
    that reference it.
  - If a value is a string constant, lift it if it is referred to only once.
  - Collapse dicts and arrays of constants to constant dicts/arrays.
  """

  def __init__(self, result: Any, values: Any | None = None):
    """Builds an ExpressionOptimizer.

    Args:
      result: The result to optimize, either as a key of "values", or as a
        quasi-Expression.
      values: If provided (in compound mode), a set of named ValueNodes.
    """
    self._result = result
    # Convert sequence of tuples to a dict allowing lookup.
    # Values will be of the form:
    #  [('0', {'constantValue': 99}), ('1', {'constantValue': 98}), ...]
    self._values = dict(values) if values is not None else None

    if self._is_compound():
      self._single_uses = self._find_single_uses()
      self._optimized_values = {}
      self._reference_map = {}

  def _is_compound(self) -> bool:
    return self._values is not None

  def _find_single_uses(self) -> set[Any]:
    """Finds the names of all named values that are referred to only once."""
    reference_counts = collections.defaultdict(int)
    reference_counts[self._result] += 1

    def _contained_reference(value: Any) -> Any | None:
      """Gets a contained reference from a ValueNode, if there is one."""
      if 'functionDefinitionValue' in value:
        return value['functionDefinitionValue']['body']
      elif 'functionInvocationValue' in value:
        function_invocation = value['functionInvocationValue']
        if 'functionReference' in function_invocation:
          return function_invocation['functionReference']
      elif 'valueReference' in value:
        return value['valueReference']
      return None

    def increment_reference_count(value: Any) -> None:
      reference = _contained_reference(value)
      if reference is not None:
        reference_counts[reference] += 1

    self._visit_all_values_in_expression(increment_reference_count)
    return {reference for reference, count in reference_counts.items()
               if count == 1}

  def optimize(self) -> Any:
    """Optimises the expression, returning the optimised form."""
    optimized_result = self._optimize_referred_value(self._result)
    if self._is_compound():
      return {'result': optimized_result, 'values': self._optimized_values}
    else:
      return optimized_result

  def _optimize_referred_value(self, reference_or_value: Any) -> Any:
    """Recursively optimises a value.

    Optimises a value and everything recursively reachable from it.

    This operates differently depending on the mode.
    In compound mode:
      Takes a name (in _values) for a ValueNode, optimises the referenced
      ValueNode, and returns a name (in _optimized_values) for the optimised
      ValueNode. Updates _optimized_values and _reference_map.
    In non-compound mode:
      Takes a quasi-ValueNode, optimises it, and returns the optimised
      quasi-ValueNode.

    Args:
      reference_or_value: The name in _values of the value to optimise, or the
        actual value itself.

    Returns:
      The name, in _optimized_values, of the optimised value, or the optimised
      value itself.
    """
    if self._is_compound():
      if reference_or_value in self._reference_map:
        return self._reference_map[reference_or_value]
      mapped_reference = str(len(self._reference_map))
      self._reference_map[reference_or_value] = mapped_reference
      self._optimized_values[mapped_reference] = self._optimize_value(
          self._values[reference_or_value], 0)
      return mapped_reference
    else:
      return self._optimize_value(reference_or_value, 0)

  def _optimize_value(self, value: Any, depth: int) -> Any:
    """Optimises a single value.

    Args:
      value: The ValueNode to optimise, in dict form.
      depth: How deep in the encoded output this value will be placed.

    Returns:
      An optimised version of that value, created by lifting in all feasible
      constants and references, subject (in compound mode) to a depth limit.
    """
    if any(
        x in value for x in
        ['constantValue', 'integerValue', 'bytesValue', 'argumentReference']):
      # Not optimisable.
      return value
    elif 'arrayValue' in value:
      # Optimise recursively, then turn an array of constants into a constant
      # array.
      optimized_array = [
          self._optimize_value(array_value, depth + 3)
          for array_value in value['arrayValue']['values']
      ]
      if all(self._is_constant_value(v) for v in optimized_array):
        optimized_array = [v['constantValue'] for v in optimized_array]
        return {'constantValue': optimized_array}
      else:
        return {'arrayValue': {'values': optimized_array}}
    elif 'dictionaryValue' in value:
      # Optimise recursively, then turn a dict of constants into a constant
      # dict.
      optimized_dict = {
          key: self._optimize_value(dict_value, depth + 3)
          for key, dict_value in value['dictionaryValue']['values'].items()
      }
      if all(self._is_constant_value(v) for v in optimized_dict.values()):
        optimized_dict = {
            k: v['constantValue'] for k, v in optimized_dict.items()
        }
        return {'constantValue': optimized_dict}
      else:
        return {'dictionaryValue': {'values': optimized_dict}}
    elif 'functionDefinitionValue' in value:
      function_definition = value['functionDefinitionValue']
      return {
          'functionDefinitionValue': {
              'argumentNames': function_definition['argumentNames'],
              'body': self._optimize_referred_value(function_definition['body'])
          }
      }
    elif 'functionInvocationValue' in value:
      function_invocation = value['functionInvocationValue']
      arguments = function_invocation['arguments']
      optimized_invocation = {}
      if 'functionName' in function_invocation:
        optimized_invocation['functionName'] = function_invocation[
            'functionName']
      else:
        optimized_invocation[
            'functionReference'] = self._optimize_referred_value(
                function_invocation['functionReference'])
      optimized_invocation['arguments'] = {
          k: self._optimize_value(v, depth + 3) for k, v in arguments.items()
      }
      return {'functionInvocationValue': optimized_invocation}
    elif 'valueReference' in value:
      # Lift if possible: anything used only here, anything lightweight.
      reference = value['valueReference']
      if not self._is_compound():
        return self._optimize_value(reference, depth)

      referenced_value = self._values[reference]
      if reference in self._single_uses and depth < _DEPTH_LIMIT:
        return self._optimize_value(referenced_value, depth)
      else:
        if self._is_always_liftable(referenced_value):
          return referenced_value
        return {'valueReference': self._optimize_referred_value(reference)}

  def _is_always_liftable(self, value: Any) -> bool:
    """Determines if a value is simple enough to lift unconditionally."""
    # Non-string constants and argument references are simple enough.
    if 'constantValue' in value:
      return self._is_liftable_constant(value['constantValue'])
    else:
      return 'argumentReference' in value

  def _is_liftable_constant(self, value: Any) -> bool:
    """Whether a constant is simple enough to lift to where it's referenced."""
    return value is None or isinstance(value, (bool, float, int))

  def _is_constant_value(self, value: Any) -> bool:
    """Whether a ValueNode (as a dict) is a constant."""
    return 'constantValue' in value

  def _visit_all_values_in_expression(self, visitor: Any) -> None:
    """Calls visitor on all ValueNodes in the expression.

    Args:
      visitor: A callable that will be invoked once at every ValueNode in the
        expression, including nested ValueNodes.
    """
    self._visit_all_values(self._result, self._values[self._result], set(),
                           visitor)

  def _visit_all_values(
      self, reference: Any, value: Any, visited: Any, visitor: Any
  ) -> None:
    """Calls visitor on a ValueNode and its descendants.

    Args:
      reference: A reference to the ValueNode, or None.
      value: The ValueNode, in dict form.
      visited: A set of references for which the visitor has already been
        invoked.
      visitor: The callable to invoke.
    """
    if reference is not None:
      if reference in visited:
        return
      visited.add(reference)
    visitor(value)

    if 'arrayValue' in value:
      for v in value['arrayValue']['values']:
        self._visit_all_values(None, v, visited, visitor)
    elif 'dictionaryValue' in value:
      d = value['dictionaryValue']['values']
      for k in sorted(d):
        self._visit_all_values(None, d[k], visited, visitor)
    elif 'functionDefinitionValue' in value:
      definition_reference = value['functionDefinitionValue']['body']
      self._visit_all_values(definition_reference,
                             self._values[definition_reference], visited,
                             visitor)
    elif 'functionInvocationValue' in value:
      function_invocation = value['functionInvocationValue']
      if 'functionReference' in function_invocation:
        function_reference = function_invocation['functionReference']
        self._visit_all_values(function_reference,
                               self._values[function_reference], visited,
                               visitor)
      arguments = function_invocation['arguments']
      for k in sorted(arguments):
        self._visit_all_values(None, arguments[k], visited, visitor)
    elif 'valueReference' in value:
      value_reference = value['valueReference']
      self._visit_all_values(value_reference, self._values[value_reference],
                             visited, visitor)



"""A deserializer that decodes EE object trees from JSON DAGs."""

import json
from typing import Any

from ee import apifunction
from ee import computedobject
from ee import customfunction
from ee import ee_date
from ee import ee_exception
from ee import encodable
from ee import function
from ee import geometry


def fromJSON(json_obj: bytes | str) -> Any:  # pylint: disable=g-bad-name
  """Deserialize an object from a JSON string appropriate for API calls.

  Args:
    json_obj: The JSON representation of the input.

  Returns:
    The deserialized object.
  """
  return decode(json.loads(json_obj))


def decode(json_obj: Any) -> Any:
  """Decodes an object previously encoded using the EE API v2 (DAG) format.

  Args:
    json_obj: The serialied object to decode.

  Returns:
    The decoded object.
  """
  if 'values' in json_obj and 'result' in json_obj:
    return decodeCloudApi(json_obj)
  named_values = {}

  # Incrementally decode scope entries if there are any.
  if isinstance(json_obj, dict) and json_obj['type'] == 'CompoundValue':
    for i, (key, value) in enumerate(json_obj['scope']):
      if key in named_values:
        raise ee_exception.EEException(
            'Duplicate scope key "%s" in scope #%d.' % (key, i))
      named_values[key] = _decodeValue(value, named_values)
    json_obj = json_obj['value']

  # Decode the final value.
  return _decodeValue(json_obj, named_values)


# pylint: disable-next=g-bad-name
def _decodeValue(json_obj: Any, named_values: dict[str, Any]) -> Any:
  """Decodes an object previously encoded using the EE API v2 (DAG) format.

  This uses a provided scope for ValueRef lookup and does not allow the
  input to be a CompoundValue.

  Args:
    json_obj: The serialied object to decode.
    named_values: The objects that can be referenced by ValueRefs.

  Returns:
    The decoded object.
  """

  # Check for primitive values.
  if (json_obj is None or isinstance(json_obj, (bool, float, int, str))):
    return json_obj

  # Check for array values.
  if isinstance(json_obj, (list, tuple)):
    return [_decodeValue(element, named_values) for element in json_obj]

  # Ensure that we've got a proper object at this point.
  if not isinstance(json_obj, dict):
    raise ee_exception.EEException(f'Cannot decode object: {json_obj}')

  # Check for explicitly typed values.
  type_name = json_obj['type']
  if type_name == 'ValueRef':
    if json_obj['value'] in named_values:
      return named_values[json_obj['value']]
    else:
      raise ee_exception.EEException(f'Unknown ValueRef: {json_obj}')
  elif type_name == 'ArgumentRef':
    var_name = json_obj['value']
    if not isinstance(var_name, str):
      raise ee_exception.EEException(f'Invalid variable name: {var_name}')
    return customfunction.CustomFunction.variable(None, var_name)
  elif type_name == 'Date':
    microseconds = json_obj['value']
    if not isinstance(microseconds, (float, int)):
      raise ee_exception.EEException(f'Invalid date value: {microseconds}')
    return ee_date.Date(microseconds / 1e3)
  elif type_name == 'Bytes':
    result = encodable.Encodable()
    result.encode = lambda encoder: json_obj
    node = {'bytesValue': json_obj['value']}
    result.encode_cloud_value = lambda encoder: node
    return result
  elif type_name == 'Invocation':
    if 'functionName' in json_obj:
      func = apifunction.ApiFunction.lookup(json_obj['functionName'])
    else:
      func = _decodeValue(json_obj['function'], named_values)
    if 'arguments' in json_obj:
      args = {key: _decodeValue(value, named_values)
                  for (key, value) in json_obj['arguments'].items()}
    else:
      args = {}
    return _invocation(func, args)
  elif type_name == 'Dictionary':
    return {key: _decodeValue(value, named_values)
                for (key, value) in json_obj['value'].items()}
  elif type_name == 'Function':
    body = _decodeValue(json_obj['body'], named_values)
    signature = {
        'name': '',
        'args': [{'name': arg_name, 'type': 'Object', 'optional': False}
                 for arg_name in json_obj['argumentNames']],
        'returns': 'Object'
    }
    return customfunction.CustomFunction(signature, lambda *args: body)
  elif type_name in ('Point', 'MultiPoint', 'LineString', 'MultiLineString',
                     'Polygon', 'MultiPolygon', 'LinearRing',
                     'GeometryCollection'):
    return geometry.Geometry(json_obj)
  elif type_name == 'CompoundValue':
    raise ee_exception.EEException('Nested CompoundValues are disallowed.')
  else:
    raise ee_exception.EEException(f'Unknown encoded object type: {type_name}')


def _invocation(func: Any, args: dict[str, Any]) -> Any:
  """Creates an EE object representing the application of `func` to `args`."""
  if isinstance(func, function.Function):
    return func.apply(args)
  elif isinstance(func, computedobject.ComputedObject):
    # We have to allow ComputedObjects for cases where invocations return a
    # function, e.g., Image.parseExpression(). These need to get turned back
    # into some kind of Function, for which we need a signature. Type
    # information has been lost at this point, so we just use ComputedObject.
    signature = {
        'name': '',
        'args': [{'name': name, 'type': 'ComputedObject', 'optional': False}
                 for name in args],
        'returns': 'ComputedObject'
    }
    return function.SecondOrderFunction(func, signature).apply(args)
  raise ee_exception.EEException(f'Invalid function value: {func}')


def fromCloudApiJSON(json_obj: str | bytes) -> Any:  # pylint: disable=g-bad-name
  """Deserializes an object from the JSON string used in Cloud API calls.

  Args:
    json_obj: The JSON representation of the input.

  Returns:
    The deserialized object.
  """
  return decodeCloudApi(json.loads(json_obj))


# pylint: disable-next=g-bad-name
def decodeCloudApi(json_obj: dict[str, Any]) -> Any:
  """Decodes an object previously encoded using the EE Cloud API format.

  Args:
    json_obj: The serialized object to decode.

  Returns:
    The decoded object.
  """

  decoded = {}
  def lookup(reference, kind):
    if reference not in decoded:
      if reference not in json_obj['values']:
        raise ee_exception.EEException(f'Cannot find {reference} {kind}')
      decoded[reference] = decode_node(json_obj['values'][reference])
    return decoded[reference]

  def decode_node(node):
    if 'constantValue' in node:
      return node['constantValue']
    elif 'arrayValue' in node:
      return [decode_node(x) for x in node['arrayValue']['values']]
    elif 'dictionaryValue' in node:
      return {
          key: decode_node(x)
          for key, x in node['dictionaryValue']['values'].items()
      }
    elif 'argumentReference' in node:
      return customfunction.CustomFunction.variable(
          None, node['argumentReference'])  # pylint: disable=protected-access
    elif 'functionDefinitionValue' in node:
      return decode_function_definition(node['functionDefinitionValue'])
    elif 'functionInvocationValue' in node:
      return decode_function_invocation(node['functionInvocationValue'])
    elif 'bytesValue' in node:
      return _decodeValue({'type': 'Bytes', 'value': node['bytesValue']}, {})
    elif 'integerValue' in node:
      return int(node['integerValue'])
    elif 'valueReference' in node:
      return lookup(node['valueReference'], 'reference')
    return None

  def decode_function_definition(
      defined: dict[str, Any],
  ) -> customfunction.CustomFunction:
    body = lookup(defined['body'], 'function body')
    signature_args = [{'name': name, 'type': 'Object', 'optional': False}
                      for name in defined['argumentNames']]
    signature = {'args': signature_args, 'name': '', 'returns': 'Object'}
    return customfunction.CustomFunction(signature, lambda *args: body)

  def decode_function_invocation(invoked: dict[str, Any]) -> Any:
    if 'functionReference' in invoked:
      func = lookup(invoked['functionReference'], 'function')
    else:
      func = apifunction.ApiFunction.lookup(invoked['functionName'])
    if 'arguments' in invoked:
      args = {key: decode_node(x) for key, x in invoked['arguments'].items()}
    else:
      args = {}
    return _invocation(func, args)

  return lookup(json_obj['result'], 'result value')



#!/usr/bin/env python3
"""Test for the ee.geometry module Point methods."""

import json
from typing import Any
from unittest import mock

import unittest
import ee
from ee import apitestcase

X = 1
Y = 2
POINT = {
    'functionInvocationValue': {
        'functionName': 'GeometryConstructors.Point',
        'arguments': {'coordinates': {'constantValue': [X, Y]}},
    }
}

X2 = X + 2
Y2 = Y + 2
POINT2 = {
    'functionInvocationValue': {
        'functionName': 'GeometryConstructors.Point',
        'arguments': {'coordinates': {'constantValue': [X2, Y2]}},
    }
}

MAX_ERROR_VAL = 1.8
MAX_ERROR = {
    'functionInvocationValue': {
        'functionName': 'ErrorMargin',
        'arguments': {'value': {'constantValue': MAX_ERROR_VAL}},
    }
}

EPSG = 'EPSG:4326'
PROJ = {
    'functionInvocationValue': {
        'functionName': 'Projection',
        'arguments': {'crs': {'constantValue': EPSG}},
    }
}


def make_expression_graph_geom(
    function_name: str, args: dict[str, Any]
) -> dict[str, Any]:
  return {
      'result': '0',
      'values': {
          '0': {
              'functionInvocationValue': {
                  'functionName': 'Geometry.' + function_name,
                  'arguments': args,
              }
          }
      },
  }


class GeometryPointTest(apitestcase.ApiTestCase):

  def setUp(self):
    super().setUp()
    self.point = ee.Geometry.Point(X, Y)
    self.point2 = ee.Geometry.Point(X2, Y2)

  def test_area(self):
    expect = make_expression_graph_geom(
        'area', {'geometry': POINT, 'maxError': MAX_ERROR, 'proj': PROJ}
    )
    actual = json.loads(self.point.area(MAX_ERROR_VAL, EPSG).serialize())
    self.assertEqual(actual, expect)

    actual = json.loads(
        self.point.area(maxError=MAX_ERROR_VAL, proj=EPSG).serialize()
    )
    self.assertEqual(actual, expect)

  def test_aside(self):
    mock_function = mock.Mock(return_value=None)
    self.assertIs(self.point, self.point.aside(mock_function))
    mock_function.assert_called_once_with(self.point)

  def test_bounds(self):
    expect = make_expression_graph_geom(
        'bounds', {'geometry': POINT, 'maxError': MAX_ERROR, 'proj': PROJ}
    )
    actual = json.loads(self.point.bounds(MAX_ERROR_VAL, EPSG).serialize())
    self.assertEqual(actual, expect)

    actual = json.loads(
        self.point.bounds(maxError=MAX_ERROR_VAL, proj=EPSG).serialize()
    )
    self.assertEqual(actual, expect)

  def test_buffer(self):
    distance = 2
    expect = make_expression_graph_geom(
        'buffer',
        {
            'distance': {'constantValue': distance},
            'geometry': POINT,
            'maxError': MAX_ERROR,
            'proj': PROJ,
        },
    )
    actual = json.loads(
        self.point.buffer(distance, MAX_ERROR_VAL, EPSG).serialize()
    )
    self.assertEqual(actual, expect)

    actual = json.loads(
        self.point.buffer(
            distance=distance, maxError=MAX_ERROR_VAL, proj=EPSG
        ).serialize()
    )
    self.assertEqual(actual, expect)

  def test_centroid(self):
    expect = make_expression_graph_geom(
        'centroid', {'geometry': POINT, 'maxError': MAX_ERROR, 'proj': PROJ}
    )
    actual = json.loads(self.point.centroid(MAX_ERROR_VAL, EPSG).serialize())
    self.assertEqual(actual, expect)

    actual = json.loads(
        self.point.centroid(maxError=MAX_ERROR_VAL, proj=EPSG).serialize()
    )
    self.assertEqual(actual, expect)

  def test_closest_point(self):
    expect = make_expression_graph_geom(
        'closestPoint',
        {'left': POINT, 'right': POINT2, 'maxError': MAX_ERROR, 'proj': PROJ},
    )
    actual = json.loads(
        self.point.closestPoint(self.point2, MAX_ERROR_VAL, EPSG).serialize()
    )
    self.assertEqual(actual, expect)

    actual = json.loads(
        self.point.closestPoint(
            right=self.point2, maxError=MAX_ERROR_VAL, proj=EPSG
        ).serialize()
    )
    self.assertEqual(actual, expect)

  def test_closest_points(self):
    expect = make_expression_graph_geom(
        'closestPoints',
        {'left': POINT, 'right': POINT2, 'maxError': MAX_ERROR, 'proj': PROJ},
    )
    actual = json.loads(
        self.point.closestPoints(self.point2, MAX_ERROR_VAL, EPSG).serialize()
    )
    self.assertEqual(actual, expect)

    actual = json.loads(
        self.point.closestPoints(
            right=self.point2, maxError=MAX_ERROR_VAL, proj=EPSG
        ).serialize()
    )
    self.assertEqual(actual, expect)

  def test_contained_in(self):
    expect = make_expression_graph_geom(
        'containedIn',
        {'left': POINT, 'right': POINT2, 'maxError': MAX_ERROR, 'proj': PROJ},
    )
    actual = json.loads(
        self.point.containedIn(self.point2, MAX_ERROR_VAL, EPSG).serialize()
    )
    self.assertEqual(actual, expect)

    actual = json.loads(
        self.point.containedIn(
            right=self.point2, maxError=MAX_ERROR_VAL, proj=EPSG
        ).serialize()
    )
    self.assertEqual(actual, expect)

  def test_contains(self):
    expect = make_expression_graph_geom(
        'contains',
        {'left': POINT, 'right': POINT2, 'maxError': MAX_ERROR, 'proj': PROJ},
    )
    actual = json.loads(
        self.point.contains(self.point2, MAX_ERROR_VAL, EPSG).serialize()
    )
    self.assertEqual(actual, expect)

    actual = json.loads(
        self.point.contains(
            right=self.point2, maxError=MAX_ERROR_VAL, proj=EPSG
        ).serialize()
    )
    self.assertEqual(actual, expect)

  def test_convex_hull(self):
    expect = make_expression_graph_geom(
        'convexHull', {'geometry': POINT, 'maxError': MAX_ERROR, 'proj': PROJ}
    )
    actual = json.loads(self.point.convexHull(MAX_ERROR_VAL, EPSG).serialize())
    self.assertEqual(actual, expect)

    actual = json.loads(
        self.point.convexHull(maxError=MAX_ERROR_VAL, proj=EPSG).serialize()
    )
    self.assertEqual(actual, expect)

  def test_coordinates(self):
    expect = make_expression_graph_geom('coordinates', {'geometry': POINT})
    actual = json.loads(self.point.coordinates().serialize())
    self.assertEqual(actual, expect)

  def test_covering_grid(self):
    scale = 8.7
    expect = make_expression_graph_geom(
        'coveringGrid',
        {'geometry': POINT, 'proj': PROJ, 'scale': {'constantValue': scale}},
    )
    actual = json.loads(self.point.coveringGrid(EPSG, scale).serialize())
    self.assertEqual(actual, expect)

    actual = json.loads(
        self.point.coveringGrid(proj=EPSG, scale=scale).serialize()
    )
    self.assertEqual(actual, expect)

  def test_cut_lines(self):
    distances = [2.2, 3.3]
    expect = make_expression_graph_geom(
        'cutLines',
        {
            'geometry': POINT,
            'distances': {'constantValue': distances},
            'maxError': MAX_ERROR,
            'proj': PROJ,
        },
    )
    actual = json.loads(
        self.point.cutLines(distances, MAX_ERROR_VAL, EPSG).serialize()
    )
    self.assertEqual(actual, expect)

    actual = json.loads(
        self.point.cutLines(
            distances=distances, maxError=MAX_ERROR_VAL, proj=EPSG
        ).serialize()
    )
    self.assertEqual(actual, expect)

  def test_difference(self):
    expect = make_expression_graph_geom(
        'difference',
        {'left': POINT, 'right': POINT2, 'maxError': MAX_ERROR, 'proj': PROJ},
    )
    actual = json.loads(
        self.point.difference(self.point2, MAX_ERROR_VAL, EPSG).serialize()
    )
    self.assertEqual(actual, expect)

    actual = json.loads(
        self.point.difference(
            right=self.point2, maxError=MAX_ERROR_VAL, proj=EPSG
        ).serialize()
    )
    self.assertEqual(actual, expect)

  def test_disjoint(self):
    expect = make_expression_graph_geom(
        'disjoint',
        {'left': POINT, 'right': POINT2, 'maxError': MAX_ERROR, 'proj': PROJ},
    )
    actual = json.loads(
        self.point.disjoint(self.point2, MAX_ERROR_VAL, EPSG).serialize()
    )
    self.assertEqual(actual, expect)

    actual = json.loads(
        self.point.disjoint(
            right=self.point2, maxError=MAX_ERROR_VAL, proj=EPSG
        ).serialize()
    )
    self.assertEqual(actual, expect)

  def test_dissolve(self):
    expect = make_expression_graph_geom(
        'dissolve', {'geometry': POINT, 'maxError': MAX_ERROR, 'proj': PROJ}
    )
    actual = json.loads(self.point.dissolve(MAX_ERROR_VAL, EPSG).serialize())
    self.assertEqual(actual, expect)

    actual = json.loads(
        self.point.dissolve(maxError=MAX_ERROR_VAL, proj=EPSG).serialize()
    )
    self.assertEqual(actual, expect)

  def test_distance(self):
    spherical = True
    expect = make_expression_graph_geom(
        'distance',
        {
            'left': POINT,
            'right': POINT2,
            'maxError': MAX_ERROR,
            'proj': PROJ,
            'spherical': {'constantValue': spherical},
        },
    )
    actual = json.loads(
        self.point.distance(
            self.point2, MAX_ERROR_VAL, EPSG, spherical
        ).serialize()
    )
    self.assertEqual(actual, expect)

    actual = json.loads(
        self.point.distance(
            right=self.point2,
            maxError=MAX_ERROR_VAL,
            proj=EPSG,
            spherical=spherical,
        ).serialize()
    )
    self.assertEqual(actual, expect)

  def test_edges_are_geodesics(self):
    expect = make_expression_graph_geom(
        'edgesAreGeodesics', {'geometry': POINT}
    )
    actual = json.loads(self.point.edgesAreGeodesics().serialize())
    self.assertEqual(actual, expect)

  def test_geodesic(self):
    expect = make_expression_graph_geom('geodesic', {'geometry': POINT})
    actual = json.loads(self.point.geodesic().serialize())
    self.assertEqual(actual, expect)

  def test_geometries(self):
    expect = make_expression_graph_geom('geometries', {'geometry': POINT})
    actual = json.loads(self.point.geometries().serialize())
    self.assertEqual(actual, expect)

  def test_intersection(self):
    expect = make_expression_graph_geom(
        'intersection',
        {'left': POINT, 'right': POINT2, 'maxError': MAX_ERROR, 'proj': PROJ},
    )
    actual = json.loads(
        self.point.intersection(self.point2, MAX_ERROR_VAL, EPSG).serialize()
    )
    self.assertEqual(actual, expect)

    actual = json.loads(
        self.point.intersection(
            right=self.point2, maxError=MAX_ERROR_VAL, proj=EPSG
        ).serialize()
    )
    self.assertEqual(actual, expect)

  def test_intersects(self):
    expect = make_expression_graph_geom(
        'intersects',
        {'left': POINT, 'right': POINT2, 'maxError': MAX_ERROR, 'proj': PROJ},
    )
    actual = json.loads(
        self.point.intersects(self.point2, MAX_ERROR_VAL, EPSG).serialize()
    )
    self.assertEqual(actual, expect)

    actual = json.loads(
        self.point.intersects(
            right=self.point2, maxError=MAX_ERROR_VAL, proj=EPSG
        ).serialize()
    )
    self.assertEqual(actual, expect)

  def test_is_unbounded(self):
    expect = make_expression_graph_geom('isUnbounded', {'geometry': POINT})
    actual = json.loads(self.point.isUnbounded().serialize())
    self.assertEqual(actual, expect)

  def test_length(self):
    expect = make_expression_graph_geom(
        'length',
        {'geometry': POINT, 'maxError': MAX_ERROR, 'proj': PROJ},
    )
    actual = json.loads(self.point.length(MAX_ERROR_VAL, EPSG).serialize())
    self.assertEqual(actual, expect)

    actual = json.loads(
        self.point.length(maxError=MAX_ERROR_VAL, proj=EPSG).serialize()
    )
    self.assertEqual(actual, expect)

  def test_perimeter(self):
    expect = make_expression_graph_geom(
        'perimeter', {'geometry': POINT, 'maxError': MAX_ERROR, 'proj': PROJ}
    )
    actual = json.loads(self.point.perimeter(MAX_ERROR_VAL, EPSG).serialize())
    self.assertEqual(actual, expect)

    actual = json.loads(
        self.point.perimeter(maxError=MAX_ERROR_VAL, proj=EPSG).serialize()
    )
    self.assertEqual(actual, expect)

  def test_projection(self):
    expect = make_expression_graph_geom('projection', {'geometry': POINT})
    actual = json.loads(self.point.projection().serialize())
    self.assertEqual(actual, expect)

  def test_simplify(self):
    expect = make_expression_graph_geom(
        'simplify', {'geometry': POINT, 'maxError': MAX_ERROR, 'proj': PROJ}
    )
    actual = json.loads(self.point.simplify(MAX_ERROR_VAL, EPSG).serialize())
    self.assertEqual(actual, expect)

    actual = json.loads(
        self.point.simplify(maxError=MAX_ERROR_VAL, proj=EPSG).serialize()
    )
    self.assertEqual(actual, expect)

  def test_symmetric_difference(self):
    expect = make_expression_graph_geom(
        'symmetricDifference',
        {'left': POINT, 'right': POINT2, 'maxError': MAX_ERROR, 'proj': PROJ},
    )
    actual = json.loads(
        self.point.symmetricDifference(
            self.point2, MAX_ERROR_VAL, EPSG
        ).serialize()
    )
    self.assertEqual(actual, expect)

    actual = json.loads(
        self.point.symmetricDifference(
            right=self.point2, maxError=MAX_ERROR_VAL, proj=EPSG
        ).serialize()
    )
    self.assertEqual(actual, expect)

  def test_transform(self):
    expect = make_expression_graph_geom(
        'transform', {'geometry': POINT, 'proj': PROJ, 'maxError': MAX_ERROR}
    )
    actual = json.loads(self.point.transform(EPSG, MAX_ERROR_VAL).serialize())
    self.assertEqual(actual, expect)

    actual = json.loads(
        self.point.transform(maxError=MAX_ERROR_VAL, proj=EPSG).serialize()
    )
    self.assertEqual(actual, expect)

  def test_type(self):
    expect = make_expression_graph_geom('type', {'geometry': POINT})
    actual = json.loads(self.point.type().serialize())
    self.assertEqual(actual, expect)

  def test_union(self):
    expect = make_expression_graph_geom(
        'union',
        {'left': POINT, 'right': POINT2, 'maxError': MAX_ERROR, 'proj': PROJ},
    )
    actual = json.loads(
        self.point.union(self.point2, MAX_ERROR_VAL, EPSG).serialize()
    )
    self.assertEqual(actual, expect)

    actual = json.loads(
        self.point.union(
            right=self.point2, maxError=MAX_ERROR_VAL, proj=EPSG
        ).serialize()
    )
    self.assertEqual(actual, expect)

  def test_within_distance(self):
    distance = 2.2
    expect = make_expression_graph_geom(
        'withinDistance',
        {
            'left': POINT,
            'right': POINT2,
            'distance': {'constantValue': distance},
            'maxError': MAX_ERROR,
            'proj': PROJ,
        },
    )
    actual = json.loads(
        self.point.withinDistance(
            self.point2, distance, MAX_ERROR_VAL, EPSG
        ).serialize()
    )
    self.assertEqual(actual, expect)

    actual = json.loads(
        self.point.withinDistance(
            right=self.point2,
            distance=distance,
            maxError=MAX_ERROR_VAL,
            proj=EPSG,
        ).serialize()
    )
    self.assertEqual(actual, expect)


if __name__ == '__main__':
  unittest.main()
